import unittest

import httpretty
import pytest
import requests
from mock import mock

import rwslib
from rwslib.rws_requests import StudyDatasetRequest
from rwslib.rwsobjects import RWSPostErrorResponse


# TODO: per the Repository, httpretty is not supporting Python3 - do we need to replace?


class TestVersion(unittest.TestCase):
    """Test for the version method"""

    @httpretty.activate
    def test_version(self):
        """A simple test, patching the get request so that it does not hit a website"""

        httpretty.register_uri(
            httpretty.GET,
            "https://innovate.mdsol.com/RaveWebServices/version",
            status=200,
            body="1.0.0",
        )

        # Now my test
        rave = rwslib.RWSConnection("https://innovate.mdsol.com")
        v = rave.send_request(rwslib.rws_requests.VersionRequest())

        self.assertEqual(v, "1.0.0")
        self.assertEqual(rave.last_result.status_code, 200)

    def test_connection_failure(self):
        """Test we get a failure if we do not retry"""
        with mock.patch("requests.sessions.Session.get") as mock_get:
            mock_get.side_effect = requests.exceptions.ConnectionError()
            rave = rwslib.RWSConnection("https://innovate.mdsol.com")
            with self.assertRaises(requests.exceptions.ConnectionError) as exc:
                v = rave.send_request(rwslib.rws_requests.VersionRequest())

    """Test with only mdsol sub-domain"""

    @httpretty.activate
    def test_sub_domain(self):
        """A simple test, patching the get request so that it does not hit a website"""

        httpretty.register_uri(
            httpretty.GET,
            "https://innovate.mdsol.com/RaveWebServices/version",
            status=200,
            body="1.0.0",
        )

        # Now my test
        rave = rwslib.RWSConnection("innovate")
        v = rave.send_request(rwslib.rws_requests.VersionRequest())

        self.assertEqual("1.0.0", v)
        self.assertEqual(rave.domain, "https://innovate.mdsol.com")
        self.assertEqual(rave.last_result.status_code, 200)

    """Test for overriding the virtual directory"""

    @httpretty.activate
    def test_virtual_directory(self):
        """A simple test, patching the get request so that it does not hit a website"""

        httpretty.register_uri(
            httpretty.GET,
            "https://innovate.mdsol.com/RWS/version",
            status=200,
            body="1.0.0",
        )

        # Now my test
        rave = rwslib.RWSConnection("https://innovate.mdsol.com", virtual_dir="RWS")
        v = rave.send_request(rwslib.rws_requests.VersionRequest())

        self.assertEqual(v, "1.0.0")
        self.assertEqual(rave.last_result.status_code, 200)

    @httpretty.activate
    def test_disable_response_preserving(self):
        """A simple test, patching the get request so that it does not hit a website"""

        httpretty.register_uri(
            httpretty.GET,
            "https://innovate.mdsol.com/RaveWebServices/version",
            status=200,
            body="1.0.0",
        )

        # Now my test
        rave = rwslib.RWSConnection("https://innovate.mdsol.com")
        v = rave.send_request(rwslib.rws_requests.VersionRequest(), preserve_last_response=False)

        self.assertEqual(v, "1.0.0")
        self.assertEqual(rave.last_result, None)


class TestMustBeRWSRequestSubclass(unittest.TestCase):
    """Test that request object passed must be RWSRequest subclass"""

    def test_basic(self):
        """Must be rwssubclass or ValueError"""

        def do():
            rave = rwslib.RWSConnection("test")
            v = rave.send_request(object())

        self.assertRaises(ValueError, do)


class TestRequestTime(unittest.TestCase):
    """Test for the last request time property"""

    @httpretty.activate
    def test_request_time(self):
        """A simple test, patching the get request so that it does not hit a website"""

        httpretty.register_uri(
            httpretty.GET,
            "https://innovate.mdsol.com/RaveWebServices/version",
            status=200,
            body="1.0.0",
        )

        # Now my test
        rave = rwslib.RWSConnection("https://innovate.mdsol.com")
        v = rave.send_request(rwslib.rws_requests.VersionRequest())
        request_time = rave.request_time
        self.assertIs(type(request_time), float)

@pytest.mark.usefixtures("double_byte_chars")
class TestErrorResponse(unittest.TestCase):
    @httpretty.activate
    def test_503_error(self):
        """Test that when we don't attempt to XMLParse a non-xml response"""

        httpretty.register_uri(
            httpretty.GET,
            "https://innovate.mdsol.com/RaveWebServices/version",
            status=503,
            body="HTTP 503 Service Temporarily Unavailable",
        )
        # Now my test
        rave = rwslib.RWSConnection("https://innovate.mdsol.com")
        with self.assertRaises(rwslib.RWSException) as exc:
            v = rave.send_request(rwslib.rws_requests.VersionRequest())
        self.assertEqual(
            "HTTP 503 Service Temporarily Unavailable", exc.exception.rws_error
        )
        self.assertEqual("Unexpected Status Code (503)", str(exc.exception))

    @httpretty.activate
    def test_500_error(self):
        """Test that when we don't attempt to XMLParse a non-xml response"""

        httpretty.register_uri(
            httpretty.GET,
            "https://innovate.mdsol.com/RaveWebServices/version",
            status=500,
            body="HTTP 500.13 Web server is too busy.",
        )
        # Now my test
        rave = rwslib.RWSConnection("https://innovate.mdsol.com")
        with self.assertRaises(rwslib.RWSException) as exc:
            v = rave.send_request(rwslib.rws_requests.VersionRequest())
        self.assertEqual("HTTP 500.13 Web server is too busy.", exc.exception.rws_error)
        self.assertEqual("Server Error (500)", str(exc.exception))

    @httpretty.activate
    def test_400_error_error_response(self):
        """Parse the IIS Response Error structure"""

        text = b"""<Response
        ReferenceNumber="5b1fa9a3-0cf3-46b6-8304-37c2e3b7d04f5"
        InboundODMFileOID="1"
        IsTransactionSuccessful = "0"
        ReasonCode="RWS00024"
        ErrorOriginLocation="/ODM/ClinicalData[1]/SubjectData[1]"
        SuccessStatistics="Rave objects touched: Subjects=0; Folders=0; Forms=0; Fields=0; LogLines=0"
        ErrorClientResponseMessage="Subject already exists.">
        </Response>
        """
        httpretty.register_uri(
            httpretty.POST,
            "https://innovate.mdsol.com/RaveWebServices/webservice.aspx?PostODMClinicalData",
            status=400,
            body=text,
        )

        # Now my test
        rave = rwslib.RWSConnection("https://innovate.mdsol.com")
        with self.assertRaises(rwslib.RWSException) as exc:
            v = rave.send_request(rwslib.rws_requests.PostDataRequest("<ODM/>"))
        self.assertEqual("Subject already exists.", str(exc.exception))
        rws_error = exc.exception.rws_error
        self.assertEqual("/ODM/ClinicalData[1]/SubjectData[1]", str(rws_error.error_origin_location))
        self.assertEqual("RWS00024", str(rws_error.reasoncode))
        self.assertTrue(isinstance(rws_error, (RWSPostErrorResponse, )))
        self.assertFalse(rws_error.istransactionsuccessful)

    @httpretty.activate
    def test_400_error_iis_error(self):
        """Test that when we don't attempt to XMLParse a non-xml response"""

        text = b"""
        <!DOCTYPE html>
        <html>
        <head>
            <meta charset="utf-8" />
            <title>OOPS! Error Occurred. Sorry about this.</title>
        </head>
        <body>
            <h2>OOPS! Error Occurred</h2>
        </body>
        </html>
        """

        httpretty.register_uri(
            httpretty.GET,
            "https://innovate.mdsol.com/RaveWebServices/version",
            status=400,
            body=text,
        )

        # Now my test
        rave = rwslib.RWSConnection("https://innovate.mdsol.com")
        with self.assertRaises(rwslib.RWSException) as exc:
            v = rave.send_request(rwslib.rws_requests.VersionRequest())
        self.assertEqual("IIS Error", str(exc.exception))

    @httpretty.activate
    def test_400_error_ODM_error(self):
        """Test that when we don't attempt to XMLParse a non-xml response"""

        text = b"""<?xml version="1.0" encoding="utf-8"?>
        <ODM xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata"
         FileType="Snapshot"
         CreationDateTime="2013-04-08T10:28:49.578-00:00"
         FileOID="4d13722a-ceb6-4419-a917-b6ad5d0bc30e"
         ODMVersion="1.3"
         mdsol:ErrorDescription="Incorrect login and password combination. [RWS00008]"
         xmlns="http://www.cdisc.org/ns/odm/v1.3" />
        """

        httpretty.register_uri(
            httpretty.GET,
            "https://innovate.mdsol.com/RaveWebServices/version",
            status=400,
            body=text,
        )

        # Now my test
        rave = rwslib.RWSConnection("https://innovate.mdsol.com")
        with self.assertRaises(rwslib.RWSException) as exc:
            v = rave.send_request(rwslib.rws_requests.VersionRequest())
        self.assertEqual(
            "Incorrect login and password combination. [RWS00008]", str(exc.exception)
        )

    @httpretty.activate
    def test_401_error_error_response_no_header(self):
        """Parse the IIS Response Error structure"""

        text = "Authorization Header not provided"

        httpretty.register_uri(
            httpretty.POST,
            "https://innovate.mdsol.com/RaveWebServices/webservice.aspx?PostODMClinicalData",
            status=401,
            body=text,
            content_type="text/html; charset=utf-8"
        )

        # Now my test
        rave = rwslib.RWSConnection("https://innovate.mdsol.com")
        with self.assertRaises(rwslib.AuthorizationException) as exc:
            v = rave.send_request(rwslib.rws_requests.PostDataRequest("<ODM/>"))
        self.assertEqual(text, str(exc.exception))

    @httpretty.activate
    def test_401_error_error_response_unauthorized(self):
        """Parse the IIS Response Error structure"""

        text = b"<h3>HTTP Error 401.0 - Unauthorized</h3>"

        httpretty.register_uri(
            httpretty.POST,
            "https://innovate.mdsol.com/RaveWebServices/webservice.aspx?PostODMClinicalData",
            status=401,
            body=text,
            content_type="text/html; charset=utf-8"
        )

        # Now my test
        rave = rwslib.RWSConnection("https://innovate.mdsol.com")
        with self.assertRaises(rwslib.RWSException) as exc:
            v = rave.send_request(rwslib.rws_requests.PostDataRequest("<ODM/>"))
        self.assertEqual("Unauthorized.", str(exc.exception))

    @httpretty.activate
    def test_401_error_error_response_unauthorized_but_wonky(self):
        """Parse the IIS Response Error structure"""

        text = b"""<Response
            ReferenceNumber="0b47fe86-542f-4070-9e7d-16396a5ef08a"
            InboundODMFileOID="Not Supplied"
            IsTransactionSuccessful="0"
            ReasonCode="RWS00092"
            ErrorClientResponseMessage="You shall not pass">
            </Response>
        """

        httpretty.register_uri(
            httpretty.POST,
            "https://innovate.mdsol.com/RaveWebServices/webservice.aspx?PostODMClinicalData",
            status=401,
            body=text,
            content_type="text/xml",
        )

        # Now my test
        rave = rwslib.RWSConnection("https://innovate.mdsol.com")
        with self.assertRaises(rwslib.RWSException) as exc:
            v = rave.send_request(rwslib.rws_requests.PostDataRequest("<ODM/>"))
        self.assertEqual("You shall not pass", str(exc.exception))

    @httpretty.activate
    def test_401_error_error_response_unauthorized_without_content_type(self):
        """Parse the IIS Response Error structure"""

        text = b"""<Response
            ReferenceNumber="0b47fe86-542f-4070-9e7d-16396a5ef08a"
            InboundODMFileOID="Not Supplied"
            IsTransactionSuccessful="0"
            ReasonCode="RWS00092"
            ErrorClientResponseMessage="You shall not pass">
            </Response>
        """

        httpretty.register_uri(
            httpretty.POST,
            "https://innovate.mdsol.com/RaveWebServices/webservice.aspx?PostODMClinicalData",
            status=401,
            body=text,
        )

        # Now my test
        rave = rwslib.RWSConnection("https://innovate.mdsol.com")
        with self.assertRaises(rwslib.RWSException) as exc:
            v = rave.send_request(rwslib.rws_requests.PostDataRequest("<ODM/>"))
        self.assertEqual("You shall not pass", str(exc.exception))

    @httpretty.activate
    def test_405_error_error_response_response_object(self):
        """Parse the IIS Response Error structure"""

        text = b"""<Response
            ReferenceNumber="0b47fe86-542f-4070-9e7d-16396a5ef08a"
            InboundODMFileOID="Not Supplied"
            IsTransactionSuccessful="0"
            ReasonCode="RWS00092"
            ErrorClientResponseMessage="You shall not pass">
            </Response>
        """

        httpretty.register_uri(
            httpretty.POST,
            "https://innovate.mdsol.com/RaveWebServices/webservice.aspx?PostODMClinicalData",
            status=405,
            body=text,
        )

        # Now my test
        rave = rwslib.RWSConnection("https://innovate.mdsol.com")
        with self.assertRaises(rwslib.RWSException) as exc:
            v = rave.send_request(rwslib.rws_requests.PostDataRequest(b"<ODM/>"))
        self.assertEqual("You shall not pass", str(exc.exception))

    @httpretty.activate
    def test_405_error_ODM_error(self):
        """Parse a 405 error represented as an ODM"""
        # NOTE: this is not a real response, 405's are handled by IIS and not
        #  put through the ODM wringer
        text = b"""\xef\xbb\xbf<?xml version="1.0" encoding="utf-8"?>
        <ODM xmlns:mdsol="http://www.mdsol.com/ns/odm/metadata"
         FileType="Snapshot"
         CreationDateTime="2013-04-08T10:28:49.578-00:00"
         FileOID="4d13722a-ceb6-4419-a917-b6ad5d0bc30e"
         ODMVersion="1.3"
         mdsol:ErrorDescription="Incorrect login and password combination. [RWS00008]"
         xmlns="http://www.cdisc.org/ns/odm/v1.3" />

        """

        httpretty.register_uri(
            httpretty.GET,
            "https://innovate.mdsol.com/RaveWebServices/studies",
            status=405,
            body=text,
            content_type="text/xml"
        )

        # Now my test
        rave = rwslib.RWSConnection("https://innovate.mdsol.com")
        with self.assertRaises(rwslib.RWSException) as exc:
            v = rave.send_request(rwslib.rws_requests.ClinicalStudiesRequest())
        self.assertEqual(
            "Incorrect login and password combination. [RWS00008]", str(exc.exception)
        )

    @httpretty.activate
    def test_405_error_iis_error(self):
        """Test we handle the IIS error page"""

        text = """
        <!DOCTYPE html>
        <html>
        <head>
            <meta charset="utf-8" />
            <title>OOPS! Error Occurred. Sorry about this.</title>
        </head>
        <body>
            <h2>OOPS! Error Occurred</h2>
        </body>
        </html>
        """

        httpretty.register_uri(
            httpretty.GET,
            "https://innovate.mdsol.com/RaveWebServices/version",
            status=405,
            body=text,
        )

        # Now my test
        rave = rwslib.RWSConnection("https://innovate.mdsol.com")
        with self.assertRaises(rwslib.RWSException) as exc:
            v = rave.send_request(rwslib.rws_requests.VersionRequest())
        self.assertEqual("Unexpected Status Code (405)", str(exc.exception))

    @httpretty.activate
    def test_multibyte_character_encoding(self):
        """
        Test the output is properly encoded, Rave sends text/xml, but the underlying library doesn't seem to detect the
            encoding correctly
        """
        httpretty.register_uri(
                httpretty.GET,
                "https://training1.mdsol.com/RaveWebServices/studies/RWS_Training_Japan(PROD)/datasets/regular/SURGERY",
                status=200,
                body=self.double_byte_chars,
                content_type="text/xml"
            )
        # Now my test
        r = rwslib.RWSConnection("https://training1.mdsol.com")
        result = r.send_request(
            StudyDatasetRequest("RWS_Training_Japan", "PROD", formoid="SURGERY")
        )
        # these don't match as the encoding of the characters fails
        self.assertNotEqual(self.double_byte_chars.decode('utf-8'), result)

        class UTF8StudyDatasetRequest(StudyDatasetRequest):
            """
            Force the encoding of the response
            """
            def result(self, response):
                response.encoding = 'utf-8'
                return response.text

        result = r.send_request(
            UTF8StudyDatasetRequest("RWS_Training_Japan", "PROD", formoid="SURGERY")
        )
        # these match as the encoding enforcement makes it so
        self.assertEqual(self.double_byte_chars.decode('utf-8'), result)


if __name__ == "__main__":
    unittest.main()




"""
Views related to clinical views and their metadata, Biostats Gateway/Adapter

See the following for more information (from the Knowledge Hub - requires login):
https://learn.medidata.com/en-US/bundle/rave-web-services/page/biostat_adapter.html

"""

from . import RWSAuthorizedGetRequest, QueryOptionGetRequest, check_dataset_type

# -----------------------------------------------------------------------------------------------------------------------
# Utilities

DATASET_FORMATS = {"csv": ".csv", "xml": ""}


def check_dataset_format(ds_format):
    """
    Ensure dataset format is XML or CSV

    :param str ds_format: Format of the Dataset (expected to be one of `csv` or `xml`)
    """
    if ds_format.lower() not in DATASET_FORMATS.keys():
        raise ValueError(
            "dataset_format is expected to be one of %s. '%s' is not valid"
            % (", ".join(DATASET_FORMATS.keys()), ds_format)
        )


def dataset_format_to_extension(ds_format):
    """
    Get the preferred Dataset format extension

    :param str ds_format: Format of the Dataset (expected to be one of `csv` or `xml`)
    :rtype: str
    """
    try:
        return DATASET_FORMATS[ds_format]
    except KeyError:
        raise ValueError(
            "dataset_format is expected to be one of %s. '%s' is not valid"
            % (", ".join(DATASET_FORMATS.keys()), ds_format)
        )


# -----------------------------------------------------------------------------------------------------------------------
# Classes


class CVMetaDataRequest(QueryOptionGetRequest):
    """Return Clinical View Metadata as ODM string"""

    KNOWN_QUERY_OPTIONS = ["versionitem", "rawsuffix", "codelistsuffix", "decodesuffix"]

    def __init__(
            self,
            project_name,
            environment_name,
            versionitem=None,
            rawsuffix=None,
            codelistsuffix=None,
            decodesuffix=None,
    ):
        """
        :param str project_name: Project Name
        :param str environment_name: Environment Name
        :param str versionitem: Adds the subject's CRF version to the dataset, and identifies it with `versionitem`
        :param str rawsuffix: Adds raw data values to a full or incremental dataset, and identifies these values
            with `rawsuffix`.
        :param str codelistsuffix: Adds code list OIDS for fields that use a code list to the dataset, and identifies
            these values with `codelistsuffix`
        :param str decodesuffix: Add decoded values of items that have an associated code list to the dataset, and
            identifies these values with `decodesuffix`.
        """
        self.project_name = project_name
        self.environment_name = environment_name

        self.versionitem = versionitem
        self.rawsuffix = rawsuffix
        self.codelistsuffix = codelistsuffix
        self.decodesuffix = decodesuffix

    def studyname_environment(self):
        """
        Combine the Study Name and Environment

        :rtype: str
        """
        return "%s(%s)" % (self.project_name, self.environment_name)

    def url_path(self):
        return self.make_url(
            "studies",
            self.studyname_environment(),
            "datasets",
            "metadata",
            "regular",
            **self._querystring()
        )


class FormDataRequest(QueryOptionGetRequest):
    """Return CV Form Data as CSV or XML"""

    KNOWN_QUERY_OPTIONS = ["start"]

    def __init__(
            self,
            project_name,
            environment_name,
            dataset_type,
            form_oid,
            start=None,
            dataset_format="csv",
    ):
        """
        :param str project_name: Project Name
        :param str environment_name: Environment Name
        :param str dataset_type: Type of dataset (either `regular` or `raw`)
        :param str form_oid: OID for the Form of interest
        :param str start: Start Date for the dataset pull (should be an iso8601 formatted date)
        :param str dataset_format: Specify format of the Datasets (either `csv` or `xml`)
        """
        check_dataset_format(dataset_format)
        self.dataset_format = dataset_format

        self.project_name = project_name
        self.environment_name = environment_name

        self.dataset_type = dataset_type
        check_dataset_type(self.dataset_type)

        self.form_oid = form_oid

        # TODO: Check start is iso8601 date/time
        self.start = start

    def studyname_environment(self):
        return "%s(%s)" % (self.project_name, self.environment_name)

    def _dataset_name(self):
        return "%s%s" % (
            self.form_oid,
            dataset_format_to_extension(self.dataset_format),
        )

    def url_path(self):
        return self.make_url(
            "studies",
            self.studyname_environment(),
            "datasets",
            self.dataset_type,
            self._dataset_name(),
            **self._querystring()
        )


class MetaDataRequest(RWSAuthorizedGetRequest):
    """Return Metadata for Clinical Views in CSV or XML fornat"""

    def __init__(self, dataset_format="csv"):
        """
        :param str dataset_format: Specify format of the Datasets (either `csv` or `xml`)
        """
        check_dataset_format(dataset_format)
        self.dataset_format = dataset_format

    def _dataset_name(self):
        return "ClinicalViewMetadata%s" % dataset_format_to_extension(
            self.dataset_format
        )

    def url_path(self):
        return self.make_url("datasets", self._dataset_name())


class ProjectMetaDataRequest(RWSAuthorizedGetRequest):
    """Return Metadata for Clinical Views in CSV or XML format for a Project"""

    def __init__(self, project_name, dataset_format="csv"):
        """
        :param str project_name: Project Name
        :param str dataset_format: Specify format of the Datasets (either `csv` or `xml`)
        """
        check_dataset_format(dataset_format)
        self.dataset_format = dataset_format.lower()
        self.project_name = project_name

    def _dataset_name(self):
        return "ClinicalViewMetadata%s" % dataset_format_to_extension(
            self.dataset_format
        )

    def url_path(self):
        return self.make_url(
            "datasets", self._dataset_name(), **{"ProjectName": self.project_name}
        )


class ViewMetaDataRequest(RWSAuthorizedGetRequest):
    """Return Metadata for Clinical Views in CSV fornat for a single View"""

    def __init__(self, view_name, dataset_format="csv"):
        """
        :param str view_name: Clinical View of interest
        :param str dataset_format: Specify format of the Datasets (either `csv` or `xml`)
        """
        check_dataset_format(dataset_format)
        self.dataset_format = dataset_format.lower()
        self.view_name = view_name

    def _dataset_name(self):
        return "ClinicalViewMetadata%s" % dataset_format_to_extension(
            self.dataset_format
        )

    def url_path(self):
        return self.make_url(
            "datasets", self._dataset_name(), **{"ViewName": self.view_name}
        )


class CommentDataRequest(RWSAuthorizedGetRequest):
    """Return Comments from Rave as CSV or XML"""

    def __init__(self, project_name, environment_name, dataset_format="csv"):
        """
        :param str project_name: Project Name
        :param str dataset_format: Specify format of the Datasets (either `csv` or `xml`)
        :param str environment_name: Environment Name
        """
        check_dataset_format(dataset_format)
        self.dataset_format = dataset_format.lower()
        self.project_name = project_name
        self.environment_name = environment_name

    def studyname_environment(self):
        return "%s(%s)" % (self.project_name, self.environment_name)

    def _dataset_name(self):
        return "SDTMComments%s" % dataset_format_to_extension(self.dataset_format)

    def url_path(self):
        return self.make_url(
            "datasets",
            self._dataset_name(),
            **{"studyid": self.studyname_environment()}
        )


class ProtocolDeviationsRequest(CommentDataRequest):
    """Retrieve Protocol Deviation Information from Rave"""

    def _dataset_name(self):
        return "SDTMProtocolDeviations%s" % dataset_format_to_extension(
            self.dataset_format
        )

    def url_path(self):
        return self.make_url(
            "datasets",
            self._dataset_name(),
            **{"studyid": self.studyname_environment()}
        )


class DataDictionariesRequest(CommentDataRequest):
    """Retrieve Data Dictionaries from Rave"""

    def _dataset_name(self):
        return "SDTMDataDictionaries%s" % dataset_format_to_extension(
            self.dataset_format
        )

    def url_path(self):
        return self.make_url(
            "datasets",
            self._dataset_name(),
            **{"studyid": self.studyname_environment()}
        )



# -*- coding: utf-8 -*-

__author__ = 'isparks'

import unittest

from rwslib.builders.metadata import *
from rwslib.builders.constants import DataType, LogicalRecordPositionType, StepType
from rwslib.builders.clinicaldata import ItemData

from ..common import obj_to_doc

# Metadata object tests


class TestTranslatedText(unittest.TestCase):
    def test_builder(self):
        """XML produced"""
        tested = TranslatedText('A test', lang='en')
        doc = obj_to_doc(tested)
        self.assertEqual(doc.tag, "TranslatedText")
        self.assertEqual("en", doc.attrib['xml:lang'])
        self.assertEqual("A test", doc.text)

    def test_builder_no_lang(self):
        """XML produced when no lang is provided"""
        tested = TranslatedText('A test')
        doc = obj_to_doc(tested)
        self.assertEqual(doc.tag, "TranslatedText")
        self.assertEqual("", doc.get('xml:lang', ''))
        self.assertEqual("A test", doc.text)

    def test_accepts_no_children(self):
        with self.assertRaises(ValueError):
            TranslatedText('test') << object()


class TestSymbols(unittest.TestCase):
    def test_can_only_receive_translated_text(self):
        with self.assertRaises(ValueError):
            Symbol() << object()

    def test_translated_text_received(self):
        tested = Symbol()
        tested << TranslatedText('en', 'Test string')
        self.assertEqual(1, len(tested.translations))

    def test_builder(self):
        """XML produced"""
        tested = Symbol()(TranslatedText('en', 'Test string'))
        doc = obj_to_doc(tested)
        self.assertEqual(doc.tag, "Symbol")
        self.assertEqual(list(doc)[0].tag, "TranslatedText")


class TestMeasurementUnits(unittest.TestCase):
    def test_basic(self):
        tested = MeasurementUnit('MU_OID', 'MU_NAME')
        self.assertEqual(1, tested.constant_a)
        self.assertEqual(1, tested.constant_b)
        self.assertEqual(0, tested.constant_c)
        self.assertEqual(0, tested.constant_k)
        self.assertEqual(0, len(tested.symbols))

    def test_kg(self):
        tested = MeasurementUnit("MSU00001", "KG", unit_dictionary_name='UN1', standard_unit=True)
        doc = obj_to_doc(tested)
        self.assertEqual(doc.attrib['mdsol:StandardUnit'], "Yes")

    def test_can_only_receive_symbol(self):
        with self.assertRaises(ValueError):
            MeasurementUnit('MU_OID', 'MU_NAME') << object()

    def test_symbol_received(self):
        tested = MeasurementUnit('MU_OID', 'MU_NAME')
        tested << Symbol()
        self.assertEqual(1, len(tested.symbols))

    def test_builder(self):
        """XML produced"""
        tested = MeasurementUnit('MU_OID', 'MU_NAME')(Symbol())
        doc = obj_to_doc(tested)
        self.assertEqual(doc.tag, "MeasurementUnit")
        self.assertEqual(list(doc)[0].tag, "Symbol")


class TestBasicDefinitions(unittest.TestCase):
    def test_basic(self):
        tested = BasicDefinitions()
        self.assertEqual(0, len(tested.measurement_units))

    def test_mus_onlu(self):
        tested = BasicDefinitions()
        with self.assertRaises(ValueError):
            tested << object()

    def test_mus(self):
        tested = BasicDefinitions()
        tested << MeasurementUnit("MU_OID", "MUNAME")
        self.assertEqual(1, len(tested.measurement_units))

    def test_builder(self):
        """XML produced"""
        tested = BasicDefinitions()(MeasurementUnit("MU_OID", "MUNAME"))
        doc = obj_to_doc(tested)
        self.assertEqual(doc.tag, "BasicDefinitions")
        self.assertEqual(list(doc)[0].tag, "MeasurementUnit")


class TestGlobalVariables(unittest.TestCase):
    """Test Global Variables class"""

    def test_basic(self):
        tested = GlobalVariables('project_name', 'name', 'description')
        self.assertEqual('project_name', tested.protocol_name)
        self.assertEqual('name', tested.name)
        self.assertEqual('description', tested.description)

    def test_defaults(self):
        tested = GlobalVariables('project_name')
        self.assertEqual('project_name', tested.protocol_name)
        self.assertEqual('project_name', tested.name)  # Defaults to protocol_name
        self.assertEqual('', tested.description)  # Defaults to empty string

    def test_no_children(self):
        """GlobalVariables accepts no children"""
        with self.assertRaises(ValueError):
            GlobalVariables('TEST') << object()

    def test_builder(self):
        """XML produced"""
        tested = GlobalVariables('project_name', description="A description")
        doc = obj_to_doc(tested)
        self.assertEqual(doc.tag, "GlobalVariables")
        self.assertEqual(list(doc)[0].tag, "StudyName")
        self.assertEqual("project_name", list(doc)[0].text)
        self.assertEqual(list(doc)[1].tag, "StudyDescription")
        self.assertEqual("A description", list(doc)[1].text)
        self.assertEqual(list(doc)[2].tag, "ProtocolName")
        self.assertEqual("project_name", list(doc)[2].text)


class TestStudyEventRef(unittest.TestCase):
    def test_cannot_accept_child(self):
        with self.assertRaises(ValueError):
            StudyEventRef("OID", 2, False) << object()

    def test_optional_order_number(self):
        ser = StudyEventRef("OID")
        doc = obj_to_doc(ser)
        self.assertIsNone(doc.get('OrderNumber'))
        self.assertEqual("No", doc.get('Mandatory'))

    def test_zero_order_number(self):
        """Not that it's entirely sensible, but it's cleaner"""
        ser = StudyEventRef("OID", order_number=0)
        doc = obj_to_doc(ser)
        self.assertEqual("0", str(doc.get('OrderNumber')))

    def test_mandatory_study_event_ref(self):
        ser = StudyEventRef("OID", mandatory=True)
        doc = obj_to_doc(ser)
        self.assertIsNone(doc.get('OrderNumber'))
        self.assertEqual("Yes", doc.get('Mandatory'))


class TestProtocol(unittest.TestCase):
    def test_can_accept_studyeventref_child(self):
        se = StudyEventRef('OID', 2, True)
        tested = Protocol()(se)
        self.assertEqual(se, tested.study_event_refs[0])

    def test_cannot_accept_non_study_eventref_child(self):
        with self.assertRaises(ValueError):
            Protocol() << object()

    def test_builder(self):
        """XML produced"""
        doc = obj_to_doc(Protocol()(StudyEventRef("OID", 1, True)))
        self.assertEqual(doc.tag, "Protocol")
        self.assertEqual(list(doc)[0].tag, "StudyEventRef")


class TestFormRef(unittest.TestCase):
    def test_cannot_accept_any_child(self):
        with self.assertRaises(ValueError):
            FormRef('OID', 1, False) << object()


class TestStudyEventDef(unittest.TestCase):
    def test_cannot_accept_non_formref_child(self):
        with self.assertRaises(ValueError):
            StudyEventDef("OID", "Name", False, StudyEventDef.SCHEDULED) << object()

    def test_can_accept_formref_child(self):
        tested = StudyEventDef("OID", "Name", False, StudyEventDef.SCHEDULED)(FormRef("FORMOID", 1, False))
        self.assertEqual(1, len(tested.formrefs))

    def test_builder(self):
        tested = StudyEventDef("OID", "Name", False,
                               StudyEventDef.SCHEDULED,
                               category="Test Cat",
                               access_days=1,
                               start_win_days=2,
                               target_days=3,
                               end_win_days=4,
                               overdue_days=5,
                               close_days=6
                               )
        tested << FormRef("FORMOID", 1, False)

        doc = obj_to_doc(tested)
        self.assertEqual(doc.tag, "StudyEventDef")
        self.assertEqual("OID", doc.attrib['OID'])
        self.assertEqual("Name", doc.attrib['Name'])
        self.assertEqual("Scheduled", doc.attrib['Type'])
        self.assertEqual("Test Cat", doc.attrib['Category'])
        self.assertEqual("1", doc.attrib['mdsol:AccessDays'])
        self.assertEqual("2", doc.attrib['mdsol:StartWinDays'])
        self.assertEqual("3", doc.attrib['mdsol:TargetDays'])
        self.assertEqual("4", doc.attrib['mdsol:EndWinDays'])
        self.assertEqual("5", doc.attrib['mdsol:OverDueDays'])
        self.assertEqual("6", doc.attrib['mdsol:CloseDays'])
        self.assertEqual("FormRef", list(doc)[0].tag)


class TestItemGroupRef(unittest.TestCase):
    def test_accepts_no_children(self):
        with self.assertRaises(ValueError):
            ItemGroupRef("ItemGroup1", 1) << object()

    def test_optional_order_number(self):
        igr = ItemGroupRef("ItemGroup1")
        doc = obj_to_doc(igr)
        self.assertEqual("ItemGroup1", doc.attrib['ItemGroupOID'])
        self.assertNotIn('OrderNumber', doc.attrib)

    def test_builder(self):
        tested = ItemGroupRef("ItemGroup1", 1, mandatory=True)
        doc = obj_to_doc(tested)
        self.assertEqual(doc.tag, "ItemGroupRef")
        self.assertEqual("ItemGroup1", doc.attrib['ItemGroupOID'])
        self.assertEqual("1", doc.attrib['OrderNumber'])
        self.assertEqual("Yes", doc.attrib['Mandatory'])


class TestMdsolHelpTexts(unittest.TestCase):
    def test_build(self):
        tested = MdsolHelpText("en", "This is a help text")
        doc = obj_to_doc(tested)
        self.assertEqual(doc.tag, "mdsol:HelpText")
        self.assertEqual("en", doc.attrib['xml:lang'])
        self.assertEqual("This is a help text", doc.text)

    def test_accepts_no_children(self):
        with self.assertRaises(ValueError):
            MdsolHelpText("en", "Content") << object()


class TestMdsolViewRestriction(unittest.TestCase):
    def test_accepts_no_children(self):
        with self.assertRaises(ValueError):
            MdsolViewRestriction("CRA") << object()

    def test_build(self):
        tested = MdsolViewRestriction("DM")
        doc = obj_to_doc(tested)
        self.assertEqual(doc.tag, "mdsol:ViewRestriction")
        self.assertEqual("DM", doc.text)


class TestMdsolEntryRestriction(unittest.TestCase):
    def test_accepts_no_children(self):
        with self.assertRaises(ValueError):
            MdsolEntryRestriction("CRA") << object()

    def test_build(self):
        tested = MdsolEntryRestriction("DM")
        doc = obj_to_doc(tested)
        self.assertEqual(doc.tag, "mdsol:EntryRestriction")
        self.assertEqual("DM", doc.text)


class TestMdsolLabelRef(unittest.TestCase):
    def test_accepts_no_children(self):
        with self.assertRaises(ValueError):
            MdsolLabelRef("LABEL1", 1) << object()

    def test_build(self):
        tested = MdsolLabelRef("OID1", 1)

        doc = obj_to_doc(tested)
        self.assertEqual(doc.tag, "mdsol:LabelRef")
        self.assertEqual("OID1", doc.attrib['LabelOID'])
        self.assertEqual("1", doc.attrib['OrderNumber'])


class TestMdsolAttribute(unittest.TestCase):
    def test_accepts_no_children(self):
        with self.assertRaises(ValueError):
            MdsolAttribute("Namespace", "Name", "Value") << object()


class TestItemRef(unittest.TestCase):
    def test_accepts_no_children(self):
        with self.assertRaises(ValueError):
            ItemRef("OID", 1) << object()

    def test_build(self):
        tested = ItemRef("OID1", 1,
                         key_sequence=3,
                         imputation_method_oid='IMPUTE1',
                         role="AROLE",
                         role_codelist_oid='ROLEX')

        tested << MdsolAttribute("Namespace", "Name", "Value")
        tested << MdsolAttribute("Namespace2", "Name2", "Value2")

        doc = obj_to_doc(tested)
        self.assertEqual(doc.tag, "ItemRef")
        self.assertEqual("OID1", doc.attrib['ItemOID'])
        self.assertEqual("1", doc.attrib['OrderNumber'])
        self.assertEqual("3", doc.attrib['KeySequence'])
        self.assertEqual("IMPUTE1", doc.attrib['ImputationMethodOID'])
        self.assertEqual("AROLE", doc.attrib['Role'])
        self.assertEqual("ROLEX", doc.attrib['RoleCodeListOID'])
        self.assertEqual("mdsol:Attribute", list(doc)[0].tag)
        self.assertEqual("mdsol:Attribute", list(doc)[1].tag)


class TestQuestion(unittest.TestCase):
    def test_cannot_accept_non_translation_child(self):
        with self.assertRaises(ValueError):
            Question() << object()

    def test_can_accept_translation_child(self):
        tested = Question()(TranslatedText("How do you feel today?"))
        self.assertEqual(1, len(tested.translations))

    def test_build(self):
        tested = Question()(TranslatedText('How are you feeling?'))
        doc = obj_to_doc(tested)
        self.assertEqual(doc.tag, "Question")
        self.assertEqual("TranslatedText", list(doc)[0].tag)


class TestMeasurementUnitRef(unittest.TestCase):
    def test_cannot_accept_child(self):
        with self.assertRaises(ValueError):
            MeasurementUnitRef("KG") << object()

    def test_build(self):
        tested = MeasurementUnitRef("KG")
        doc = obj_to_doc(tested)
        self.assertEqual(doc.tag, "MeasurementUnitRef")
        self.assertEqual("KG", doc.attrib['MeasurementUnitOID'])


class TestCodeListRef(unittest.TestCase):
    def test_cannot_accept_child(self):
        with self.assertRaises(ValueError):
            CodeListRef("SEV") << object()

    def test_build(self):
        tested = CodeListRef("SEVERITY")
        doc = obj_to_doc(tested)
        self.assertEqual(doc.tag, "CodeListRef")
        self.assertEqual("SEVERITY", doc.attrib['CodeListOID'])


class TestMdsolReviewGroup(unittest.TestCase):
    def test_accepts_no_children(self):
        with self.assertRaises(ValueError):
            MdsolReviewGroup("DM") << object()


    def test_build(self):
        doc = obj_to_doc(MdsolReviewGroup("CRA"))
        self.assertEqual(doc.tag, "mdsol:ReviewGroup")
        self.assertEqual("CRA", doc.text)


class TestMdsolLabelDef(unittest.TestCase):
    def setUp(self):
        self.tested = MdsolLabelDef("L_AGE", "Age Label", field_number=4)

    def test_accepts_no_strange_children(self):
        with self.assertRaises(ValueError):
            self.tested << object()

    def test_accepts_view_restriction(self):
        self.tested << MdsolViewRestriction("DM")
        self.assertEqual(1, len(self.tested.view_restrictions))

    def test_accepts_translated_text(self):
        tt = TranslatedText("Hello")
        self.tested << tt
        self.assertEqual(tt, self.tested.translations[0])

    def test_build(self):
        self.tested << TranslatedText("Please answer all questions")
        self.tested << MdsolViewRestriction("DM")

        doc = obj_to_doc(self.tested)
        self.assertEqual(doc.tag, "mdsol:LabelDef")
        self.assertEqual("L_AGE", doc.attrib['OID'])
        self.assertEqual("Age Label", doc.attrib['Name'])
        self.assertEqual("4", doc.attrib['FieldNumber'])
        self.assertEqual("TranslatedText", list(doc)[0].tag)
        self.assertEqual("mdsol:ViewRestriction", list(doc)[1].tag)


class TestCheckValue(unittest.TestCase):
    def test_accepts_no_children(self):
        with self.assertRaises(ValueError):
            CheckValue("Test") << object()

    def test_build(self):
        doc = obj_to_doc(CheckValue(99))
        self.assertEqual(doc.tag, "CheckValue")
        self.assertEqual("99", doc.text)


class TestRangeCheck(unittest.TestCase):
    def test_accepts_no_strange_children(self):
        with self.assertRaises(ValueError):
            RangeCheck(comparator=RangeCheckComparatorType.GreaterThanEqualTo,
                       soft_hard=RangeCheckType.Soft) << object()

    def test_accepts_no_strange_soft_hard(self):
        with self.assertRaises(AttributeError):
            RangeCheck(comparator=RangeCheckComparatorType.GreaterThanEqualTo, soft_hard="Blash")

    def test_accepts_no_strange_comparator(self):
        with self.assertRaises(AttributeError):
            RangeCheck(comparator="EQ", soft_hard="Blash")

    def test_accepts_checkvalue(self):
        tested = RangeCheck(comparator=RangeCheckComparatorType.LessThanEqualTo, soft_hard=RangeCheckType.Soft)
        cv = CheckValue(0)
        tested << cv
        self.assertEqual(cv, tested.check_value)

    def test_accepts_measurement_unit_ref(self):
        tested = RangeCheck(comparator=RangeCheckComparatorType.GreaterThanEqualTo, soft_hard=RangeCheckType.Soft)
        mr = MeasurementUnitRef('kg')
        tested << mr
        self.assertEqual(mr, tested.measurement_unit_ref)

    def test_build(self):
        self.tested = RangeCheck(comparator=RangeCheckComparatorType.GreaterThanEqualTo, soft_hard=RangeCheckType.Soft)
        self.tested << CheckValue(0)
        self.tested << MeasurementUnitRef('kg')

        doc = obj_to_doc(self.tested)
        self.assertEqual("RangeCheck", doc.tag)
        self.assertEqual("Soft", doc.attrib['SoftHard'])
        self.assertEqual("GE", doc.attrib['Comparator'])
        self.assertEqual("CheckValue", list(doc)[0].tag)
        self.assertEqual("MeasurementUnitRef", list(doc)[1].tag)


class TestMdsolHeaderText(unittest.TestCase):
    def test_lang_default(self):
        tested = MdsolHeaderText("Content", "en")
        self.assertEqual("en", tested.lang)
        doc = obj_to_doc(tested)
        self.assertEqual("mdsol:HeaderText", doc.tag)
        self.assertEqual("Content", doc.text)
        self.assertEqual("en", doc.attrib['xml:lang'])


class TestItemDef(unittest.TestCase):
    def setUp(self):
        self.tested = ItemDef("I_AGE", "Age", DataType.Integer, 3,
                              significant_digits=3,
                              sas_field_name='SAGE',
                              sds_var_name='SVARNAME',
                              sas_format="3.0",
                              sas_label='AGE_YRS',
                              source_document_verify=True,
                              query_future_date=False,
                              visible=True,
                              translation_required=True,
                              query_non_conformance=True,
                              other_visits=False,
                              can_set_item_group_date=False,
                              can_set_form_date=False,
                              can_set_study_event_date=False,
                              can_set_subject_date=False,
                              visual_verify=True,
                              does_not_break_signature=True,
                              acceptable_file_extensions='jpg',
                              control_type=ControlType.Text,
                              variable_oid='SOMETHING_DIFFERENT',
                              default_value=99,
                              origin='An origin',
                              comment='A comment',
                              date_time_format='mmm yy dd',
                              field_number='10'
                              )

    def test_accepts_no_strange_children(self):
        with self.assertRaises(ValueError):
            self.tested << object()

    def test_invalid_datatype(self):
        with self.assertRaises(AttributeError):
            ItemDef("TEST", "My Test", "TOTALLY_WRONG_DATATYPE", 10)

    def test_invalid_controltype(self):
        with self.assertRaises(AttributeError):
            ItemDef("TEST", "My Test", DataType.Text, 10, control_type="TOTALLY_WRONG_CONTROLTYPE")

    def test_invalid_integer_missing_length(self):
        """Test that integer type raises error if no length set"""
        with self.assertRaises(AttributeError):
            ItemDef("TEST", "My Test", DataType.Integer)

    def test_invalid_text_missing_length(self):
        """Test that text type raises error if no length set"""
        with self.assertRaises(AttributeError):
            ItemDef("TEST", "My Test", DataType.Text)

    def test_valid_date_missing_length(self):
        """Test that dates are defaulted to length of their format"""
        id = ItemDef("TEST", "My Test", DataType.Date, date_time_format="mmm dd yy")
        self.assertEqual(9, id.length)

    def test_accepts_mdsolhelp(self):
        self.tested << MdsolHelpText("en", "Content of help")
        self.assertEqual(1, len(self.tested.help_texts))

    def test_accepts_view_restriction(self):
        self.tested << MdsolViewRestriction("DM")
        self.assertEqual(1, len(self.tested.view_restrictions))

    def test_accepts_header_text(self):
        ht = MdsolHeaderText("KILOs")
        self.tested << ht
        self.assertEqual(ht, self.tested.header_text)

    def test_does_not_accept_two_header_texts(self):
        h1 = MdsolHeaderText("One")
        h2 = MdsolHeaderText("Two")
        self.tested << h1
        with self.assertRaises(ValueError):
            self.tested << h2

    def test_accepts_measurement_unit_ref(self):
        mu = MeasurementUnitRef("KG")
        self.tested << mu
        self.assertEqual(1, len(self.tested.measurement_unit_refs))

    def test_accepts_question(self):
        q = Question()
        self.tested << q
        self.assertEqual(q, self.tested.question)

    def test_does_not_accept_two_questions(self):
        q1 = Question()
        q2 = Question()
        self.tested << q1
        with self.assertRaises(ValueError):
            self.tested << q2

    def test_accepts_codelistref(self):
        cl = CodeListRef("SEVERITY")
        self.tested << cl
        self.assertEqual(cl, self.tested.codelistref)

    def test_does_not_accept_two_codelists(self):
        cl1 = CodeListRef("SEVERITY")
        cl2 = CodeListRef("STATUS")
        self.tested << cl1
        with self.assertRaises(ValueError):
            self.tested << cl2

    def test_accepts_entry_restriction(self):
        self.tested << MdsolEntryRestriction("CRA")
        self.assertEqual(1, len(self.tested.entry_restrictions))

    def test_accepts_review_group(self):
        self.tested << MdsolReviewGroup("CRA")
        self.assertEqual(1, len(self.tested.review_groups))

    def test_accepts_range_check(self):
        self.tested << RangeCheck(RangeCheckComparatorType.LessThanEqualTo, RangeCheckType.Soft)
        self.assertEqual(1, len(self.tested.range_checks))

    def test_build(self):
        self.tested << Question()(TranslatedText("How do you feel today?"))
        self.tested << CodeListRef("SCALE_1")
        self.tested << MeasurementUnitRef("Years")
        self.tested << MdsolHelpText("en", "Content of help")
        self.tested << MdsolViewRestriction("DM")
        self.tested << MdsolEntryRestriction("CRA")
        self.tested << MdsolHeaderText("YRS")
        self.tested << MdsolReviewGroup("CRA")
        self.tested << RangeCheck(RangeCheckComparatorType.LessThanEqualTo, RangeCheckType.Soft)

        doc = obj_to_doc(self.tested)
        self.assertEqual(doc.tag, "ItemDef")
        self.assertEqual("I_AGE", doc.attrib['OID'])
        self.assertEqual("Age", doc.attrib['Name'])
        self.assertEqual("Yes", doc.attrib['mdsol:Active'])
        self.assertEqual("integer", doc.attrib['DataType'])
        self.assertEqual("3", doc.attrib['Length'])
        self.assertEqual("Text", doc.attrib['mdsol:ControlType'])
        self.assertEqual("3", doc.attrib['SignificantDigits'])
        self.assertEqual("SAGE", doc.attrib['SASFieldName'])
        self.assertEqual("SVARNAME", doc.attrib['SDSVarName'])
        self.assertEqual("AGE_YRS", doc.attrib['mdsol:SASLabel'])
        self.assertEqual("3.0", doc.attrib['mdsol:SASFormat'])
        self.assertEqual("A comment", doc.attrib['Comment'])
        self.assertEqual("An origin", doc.attrib['Origin'])
        self.assertEqual("No", doc.attrib['mdsol:QueryFutureDate'])
        self.assertEqual("Yes", doc.attrib['mdsol:Visible'])
        self.assertEqual("Yes", doc.attrib['mdsol:TranslationRequired'])
        self.assertEqual("Yes", doc.attrib['mdsol:SourceDocument'])
        self.assertEqual("No", doc.attrib['mdsol:OtherVisits'])
        self.assertEqual("Yes", doc.attrib['mdsol:SourceDocument'])
        self.assertEqual("Yes", doc.attrib['mdsol:QueryNonConformance'])
        self.assertEqual("No", doc.attrib['mdsol:CanSetItemGroupDate'])
        self.assertEqual("No", doc.attrib['mdsol:CanSetFormDate'])
        self.assertEqual("No", doc.attrib['mdsol:CanSetStudyEventDate'])
        self.assertEqual("No", doc.attrib['mdsol:CanSetSubjectDate'])
        self.assertEqual("Yes", doc.attrib['mdsol:VisualVerify'])
        self.assertEqual("Yes", doc.attrib['mdsol:DoesNotBreakSignature'])
        self.assertEqual("SOMETHING_DIFFERENT", doc.attrib['mdsol:VariableOID'])
        self.assertEqual("jpg", doc.attrib['mdsol:AcceptableFileExtensions'])
        self.assertEqual("99", doc.attrib['mdsol:DefaultValue'])
        self.assertEqual("mmm yy dd", doc.attrib['mdsol:DateTimeFormat'])
        self.assertEqual("10", doc.attrib['mdsol:FieldNumber'])

        self.assertEqual("Question", list(doc)[0].tag)
        self.assertEqual("CodeListRef", list(doc)[1].tag)
        self.assertEqual("MeasurementUnitRef", list(doc)[2].tag)
        self.assertEqual("RangeCheck", list(doc)[3].tag)
        self.assertEqual("mdsol:HeaderText", list(doc)[4].tag)
        self.assertEqual("mdsol:ViewRestriction", list(doc)[5].tag)
        self.assertEqual("mdsol:EntryRestriction", list(doc)[6].tag)
        self.assertEqual("mdsol:HelpText", list(doc)[7].tag)
        self.assertEqual("mdsol:ReviewGroup", list(doc)[8].tag)


class TestItemGroupDef(unittest.TestCase):
    def test_does_notaccept_non_itemref_child(self):
        with self.assertRaises(ValueError):
            ItemGroupDef("DM", "Demog") << object()

    def test_accepts_itemref_child(self):
        tested = ItemGroupDef("DM", "Demog")
        tested << ItemRef("OID1", "1")
        tested << MdsolLabelRef("LABEL1", 2)

        self.assertEqual(1, len(tested.item_refs))
        self.assertEqual(1, len(tested.label_refs))

    def test_build(self):
        tested = ItemGroupDef("DM", "Demography",
                              repeating=True,
                              is_reference_data=True,
                              sas_dataset_name='DMSAS',
                              domain='TESTDOMAIN',
                              origin='TESTORIGIN',
                              role='TESTROLE',
                              purpose='TESTPURPOSE',
                              comment='A comment')

        tested << ItemRef("OID1", 1)
        tested << MdsolLabelRef("LABEL1", 2)

        doc = obj_to_doc(tested)
        self.assertEqual(doc.tag, "ItemGroupDef")
        self.assertEqual("DM", doc.attrib['OID'])
        self.assertEqual("Yes", doc.attrib['Repeating'])
        self.assertEqual("Yes", doc.attrib['IsReferenceData'])
        self.assertEqual("Demography", doc.attrib['Name'])
        self.assertEqual("DMSAS", doc.attrib['SASDatasetName'])
        self.assertEqual("TESTDOMAIN", doc.attrib['Domain'])
        self.assertEqual("TESTORIGIN", doc.attrib['Origin'])
        self.assertEqual("TESTROLE", doc.attrib['Role'])
        self.assertEqual("TESTPURPOSE", doc.attrib['Purpose'])
        self.assertEqual("A comment", doc.attrib['Comment'])
        self.assertEqual("ItemRef", list(doc)[0].tag)
        self.assertEqual("mdsol:LabelRef", list(doc)[1].tag)


class TestFormDef(unittest.TestCase):
    def test_only_accept_itemgroup_ref(self):
        with self.assertRaises(ValueError):
            FormDef("VS", "Vital Signs") << object()

    def test_accept_itemgroup_ref(self):
        tested = FormDef("DM", "Demog", order_number=1)(ItemGroupRef("ItemGroup1", 1))
        self.assertEqual(1, len(tested.itemgroup_refs))

    def test_builder(self):
        tested = FormDef("DM", "Demog", repeating=True,
                         order_number=2,
                         link_form_oid='FRM1',
                         link_study_event_oid='EVT1')

        tested << ItemGroupRef("ItemGroup1", 1)
        tested << MdsolHelpText("en", "This is a help text")
        tested << MdsolViewRestriction("DM")
        tested << MdsolEntryRestriction("CRA")

        doc = obj_to_doc(tested)
        self.assertEqual(doc.tag, "FormDef")
        self.assertEqual("DM", doc.attrib['OID'])
        self.assertEqual("Demog", doc.attrib['Name'])
        self.assertEqual("Yes", doc.attrib['Repeating'])
        self.assertEqual("2", doc.attrib['mdsol:OrderNumber'])
        # Would not see LinkFormOID and LinkStudyEventOID together, they are mutually exclusive. Just for coverage.
        self.assertEqual("FRM1", doc.attrib['mdsol:LinkFormOID'])
        self.assertEqual("EVT1", doc.attrib['mdsol:LinkStudyEventOID'])
        self.assertEqual("ItemGroupRef", list(doc)[0].tag)
        self.assertEqual("mdsol:HelpText", list(doc)[1].tag)
        self.assertEqual("mdsol:ViewRestriction", list(doc)[2].tag)
        self.assertEqual("mdsol:EntryRestriction", list(doc)[3].tag)


class TestDecode(unittest.TestCase):
    def test_cannot_accept_non_translated_text(self):
        with self.assertRaises(ValueError):
            Decode() << object()

    def test_accepts_decode(self):
        tested = Decode()
        tt = TranslatedText("Yes")
        tested.add(tt)
        self.assertEqual(tt, tested.translations[0])

    def test_builder(self):
        """XML produced"""
        tested = Decode()
        tested << TranslatedText("Yes")
        doc = obj_to_doc(tested)
        self.assertEqual("Decode", doc.tag)
        self.assertEqual("TranslatedText", list(doc)[0].tag)


class TestCodeListItem(unittest.TestCase):
    def test_cannot_accept_non_decode(self):
        with self.assertRaises(ValueError):
            CodeListItem("M") << object()

    def test_accepts_decode(self):
        tested = CodeListItem("N")
        decode = Decode()
        tested.add(decode)
        self.assertEqual(decode, tested.decode)

    def test_builder_basic(self):
        """XML produced"""
        tested = CodeListItem("Y")
        tested << Decode()
        doc = obj_to_doc(tested)
        self.assertEqual("CodeListItem", doc.tag)
        self.assertEqual("", doc.get('mdsol:Specify', ''))
        self.assertEqual("", doc.get('mdsol:OrderNumber', ''))
        self.assertEqual("Y", doc.attrib['CodedValue'])
        self.assertEqual("Decode", list(doc)[0].tag)

    def test_builder_order_specify(self):
        """XML produced with optional params set"""
        tested = CodeListItem("Y", order_number=1, specify=True)
        tested << Decode()
        doc = obj_to_doc(tested)
        self.assertEqual("CodeListItem", doc.tag)
        self.assertEqual("Yes", doc.attrib['mdsol:Specify'])
        self.assertEqual("1", doc.attrib['mdsol:OrderNumber'])
        self.assertEqual("Y", doc.attrib['CodedValue'])
        self.assertEqual("Decode", list(doc)[0].tag)


class TestCodeList(unittest.TestCase):
    """Codelists contain codelistitems"""

    def test_cannot_accept_non_codelistitem(self):
        with self.assertRaises(ValueError):
            CodeList("CL1", "Codelist1", DataType.Integer) << object()

    def test_invalid_datatype(self):
        with self.assertRaises(ValueError):
            CodeList("CL1", "Codelist1", "IncorrectDataType")

    def test_accepts_codelistitem(self):
        tested = CodeList("CL1", "Codelist1", DataType.Integer)
        cl1 = CodeListItem("1")
        tested.add(cl1)
        self.assertEqual(cl1, tested.codelist_items[0])

    def test_builder(self):
        """XML produced"""
        tested = CodeList("CL_YN", "YesNo", DataType.String, sas_format_name="YESNO_CL")
        tested << CodeListItem("Y")
        doc = obj_to_doc(tested)
        self.assertEqual("CodeList", doc.tag)
        self.assertEqual(DataType.String.value, doc.attrib['DataType'])
        self.assertEqual("YESNO_CL", doc.attrib['SASFormatName'])
        self.assertEqual("CodeListItem", list(doc)[0].tag)


class TestConfirmationMessage(unittest.TestCase):
    def test_build(self):
        cm = MdsolConfirmationMessage("Form saved.")
        doc = obj_to_doc(cm)
        self.assertEqual("mdsol:ConfirmationMessage", doc.tag)
        self.assertEqual("Form saved.", doc.text)

    def test_lang_set(self):
        cm = MdsolConfirmationMessage("Form saved.", lang="en")
        doc = obj_to_doc(cm)
        self.assertEqual("en", doc.attrib['xml:lang'])


class TestMdsolCheckAction(unittest.TestCase):
    """Test extensions to ODM for Edit Checks Steps in Rave"""

    def test_build(self):
        tested = MdsolCheckAction(
            field_oid="Field",
            variable_oid="TheVAR",
            form_oid="The Form",
            form_repeat_number=1,
            folder_oid="The Folder",
            folder_repeat_number=2,
            record_position=3,
            check_action_type=ActionType.SetDataPoint,
            check_string="CString",  # Knowing which of these to use for each action type is
            check_script="CScript",  # the trick. Best to look in an Architect Loader Spreadsheet
            check_options="COptions")  # to get an idea
        doc = obj_to_doc(tested)
        self.assertEqual("mdsol:CheckAction", doc.tag)
        self.assertEqual("The Folder", doc.attrib['FolderOID'])
        self.assertEqual("The Form", doc.attrib['FormOID'])
        self.assertEqual("Field", doc.attrib['FieldOID'])
        self.assertEqual("TheVAR", doc.attrib['VariableOID'])
        self.assertEqual("1", doc.attrib['FormRepeatNumber'])
        self.assertEqual("2", doc.attrib['FolderRepeatNumber'])
        self.assertEqual("3", doc.attrib['RecordPosition'])
        self.assertEqual("CString", doc.attrib['String'])
        self.assertEqual("CScript", doc.attrib['Script'])
        self.assertEqual("COptions", doc.attrib['Options'])
        self.assertEqual(ActionType.SetDataPoint.value, doc.attrib['Type'])

    def test_invalid_action(self):
        with self.assertRaises(AttributeError):
            MdsolCheckAction(check_action_type='bad_name')


class TestMdsolCheckStep(unittest.TestCase):
    """Test extensions to ODM for Edit Checks Steps in Rave"""

    def test_build(self):
        tested = MdsolCheckStep(data_format="$1", static_value="1")
        doc = obj_to_doc(tested)
        self.assertEqual("mdsol:CheckStep", doc.tag)
        self.assertEqual("$1", doc.attrib['DataFormat'])
        self.assertEqual("1", doc.attrib['StaticValue'])
        # No function param
        self.assertEqual("", doc.attrib.get('Function', ''))

    def test_build_function(self):
        tested = MdsolCheckStep(function=StepType.Add)
        doc = obj_to_doc(tested)
        self.assertEqual("mdsol:CheckStep", doc.tag)
        self.assertEqual(StepType.Add.value, doc.attrib['Function'])
        # No data format param
        self.assertEqual("", doc.attrib.get('DataFormat', ''))

    def test_build_datastep(self):
        tested = MdsolCheckStep(variable_oid="VAROID", field_oid="FIELDOID",
                                form_oid="MyForm",
                                folder_oid="MyFolder",
                                record_position=0, form_repeat_number=2, folder_repeat_number=3,
                                logical_record_position=LogicalRecordPositionType.MaxBySubject)
        doc = obj_to_doc(tested)
        self.assertEqual("mdsol:CheckStep", doc.tag)
        self.assertEqual("VAROID", doc.attrib['VariableOID'])
        self.assertEqual("FIELDOID", doc.attrib['FieldOID'])
        self.assertEqual("MyForm", doc.attrib['FormOID'])
        self.assertEqual("MyFolder", doc.attrib['FolderOID'])
        self.assertEqual("0", doc.attrib['RecordPosition'])
        self.assertEqual("2", doc.attrib['FormRepeatNumber'])
        self.assertEqual("3", doc.attrib['FolderRepeatNumber'])
        self.assertEqual("MaxBySubject", doc.attrib['LogicalRecordPosition'])
        # No data format param
        self.assertEqual("", doc.attrib.get('DataFormat', ''))
        # No function param
        self.assertEqual("", doc.attrib.get('Function', ''))

    def test_build_custom_function(self):
        tested = MdsolCheckStep(custom_function="AlwaysTrue*")
        doc = obj_to_doc(tested)
        self.assertEqual("mdsol:CheckStep", doc.tag)
        self.assertEqual("AlwaysTrue*", doc.attrib['CustomFunction'])
        # No data format param
        self.assertEqual("", doc.attrib.get('DataFormat', ''))
        # No function param
        self.assertEqual("", doc.attrib.get('Function', ''))

    def test_invalid_function(self):
        with self.assertRaises(AttributeError):
            MdsolCheckStep(function='bad_name')

    def test_create_with_valid_lrp(self):
        """We create a function with a valid LRP value"""
        tested = MdsolCheckStep(custom_function="AlwaysTrue*",
                                logical_record_position=LogicalRecordPositionType.Last)
        doc = obj_to_doc(tested)
        self.assertEqual("mdsol:CheckStep", doc.tag)
        self.assertEqual("AlwaysTrue*", doc.attrib['CustomFunction'])
        self.assertEqual("Last", doc.attrib['LogicalRecordPosition'])

    def test_create_with_invalid_lrp(self):
        """We create a function with an invalid LRP value"""
        with self.assertRaises(AttributeError) as exc:
            tested = MdsolCheckStep(custom_function="AlwaysTrue*",
                                    logical_record_position='Wibble')
        self.assertEqual("Invalid Check Step Logical Record Position Wibble",
                         str(exc.exception))


class TestMdsolEditCheckDef(unittest.TestCase):
    """Test extensions to ODM for Edit Checks in Rave"""

    def test_build(self):
        tested = MdsolEditCheckDef("CHECK1")

        tested << MdsolCheckStep(data_format="$1", static_value="1")
        tested << MdsolCheckAction()

        doc = obj_to_doc(tested)
        self.assertEqual("mdsol:EditCheckDef", doc.tag)
        self.assertEqual("TRUE", doc.attrib['Active'])
        self.assertEqual("CHECK1", doc.attrib['OID'])
        self.assertEqual("FALSE", doc.attrib['NeedsRetesting'])
        self.assertEqual("FALSE", doc.attrib['BypassDuringMigration'])
        self.assertEqual("mdsol:CheckStep", list(doc)[0].tag)
        self.assertEqual("mdsol:CheckAction", list(doc)[1].tag)

    def test_cannot_accept_non_check_or_action_child(self):
        with self.assertRaises(ValueError):
            MdsolEditCheckDef("OID") << object()

    def test_accepts_check_step(self):
        tested = MdsolEditCheckDef("CHECK1")
        cs = MdsolCheckStep(data_format="$1", static_value="1")
        tested << cs
        self.assertEqual(cs, tested.check_steps[0])

    def test_accepts_check_action(self):
        tested = MdsolEditCheckDef("CHECK1")
        ca = MdsolCheckAction()
        tested << ca
        self.assertEqual(ca, tested.check_actions[0])


class TestMdsolDerivationStep(unittest.TestCase):
    """Test extensions to ODM for Derivation Steps in Rave"""

    def test_build(self):
        """
        We build a Derivation Value Step
        """
        tested = MdsolDerivationStep(data_format="$1", value="1")
        doc = obj_to_doc(tested)
        self.assertEqual("mdsol:DerivationStep", doc.tag)
        self.assertEqual("$1", doc.attrib['DataFormat'])
        self.assertEqual("1", doc.attrib['Value'])
        # No function param
        self.assertEqual("", doc.attrib.get('Function', ''))

    def test_build_function(self):
        """We build a Derivation Function Step"""
        tested = MdsolDerivationStep(function=StepType.Add)
        doc = obj_to_doc(tested)
        self.assertEqual("mdsol:DerivationStep", doc.tag)
        self.assertEqual(StepType.Add.value, doc.attrib['Function'])
        # No data format param
        self.assertEqual("", doc.attrib.get('DataFormat', ''))

    def test_build_datastep(self):
        """We build a Derivation Data Step"""
        tested = MdsolDerivationStep(variable_oid="VAROID", field_oid="FIELDOID",
                                     form_oid="VFORM",
                                     folder_oid="MyFolder",
                                     record_position=0, form_repeat_number=2, folder_repeat_number=3,
                                     logical_record_position=LogicalRecordPositionType.MaxBySubject)
        doc = obj_to_doc(tested)
        self.assertEqual("mdsol:DerivationStep", doc.tag)
        self.assertEqual("VAROID", doc.attrib['VariableOID'])
        self.assertEqual("FIELDOID", doc.attrib['FieldOID'])
        self.assertEqual("VFORM", doc.attrib['FormOID'])
        self.assertEqual("MyFolder", doc.attrib['FolderOID'])
        self.assertEqual("0", doc.attrib['RecordPosition'])
        self.assertEqual("2", doc.attrib['FormRepeatNumber'])
        self.assertEqual("3", doc.attrib['FolderRepeatNumber'])
        self.assertEqual("MaxBySubject", doc.attrib['LogicalRecordPosition'])
        # No data format param
        self.assertEqual("", doc.attrib.get('DataFormat', ''))
        # No function param
        self.assertEqual("", doc.attrib.get('Function', ''))

    def test_build_custom_function(self):
        """We build a Custom Function Step"""
        tested = MdsolDerivationStep(custom_function="AlwaysTrue*")
        doc = obj_to_doc(tested)
        self.assertEqual("mdsol:DerivationStep", doc.tag)
        self.assertEqual("AlwaysTrue*", doc.attrib['CustomFunction'])
        # No data format param
        self.assertEqual("", doc.attrib.get('DataFormat', ''))
        # No function param
        self.assertEqual("", doc.attrib.get('Function', ''))

    def test_invalid_function(self):
        """Trying to create a Derivation with an invalid step raises an exception"""
        with self.assertRaises(AttributeError):
            # StepType.IsPresent not valid for DerivationStep
            MdsolDerivationStep(function=StepType.IsPresent)

    def test_create_with_valid_lrp(self):
        """We create a function with a valid LRP value"""
        tested = MdsolDerivationStep(custom_function="AlwaysTrue*",
                                     logical_record_position=LogicalRecordPositionType.Last)
        doc = obj_to_doc(tested)
        self.assertEqual("mdsol:DerivationStep", doc.tag)
        self.assertEqual("AlwaysTrue*", doc.attrib['CustomFunction'])
        self.assertEqual("Last", doc.attrib['LogicalRecordPosition'])

    def test_create_with_invalid_lrp(self):
        """We create a function with an invalid LRP value"""
        with self.assertRaises(AttributeError) as exc:
            tested = MdsolDerivationStep(custom_function="AlwaysTrue*",
                                         logical_record_position='Wibble')
        self.assertEqual("Invalid Derivation Logical Record Position Wibble",
                         str(exc.exception))


class TestMdsolDerivationDef(unittest.TestCase):
    """Test extensions to ODM for Derivations in Rave"""

    def test_build(self):
        tested = MdsolDerivationDef("AGE",
                                    variable_oid="VAROID", field_oid="FIELDOID",
                                    form_oid="MyForm",
                                    folder_oid="MyFolder",
                                    record_position=0, form_repeat_number=2, folder_repeat_number=3,
                                    logical_record_position=LogicalRecordPositionType.MaxBySubject,
                                    all_variables_in_fields=True,
                                    all_variables_in_folders=True)
        doc = obj_to_doc(tested)
        tested << MdsolDerivationStep(function=StepType.Age)

        doc = obj_to_doc(tested)
        self.assertEqual("mdsol:DerivationDef", doc.tag)
        self.assertEqual("TRUE", doc.attrib['Active'])
        self.assertEqual("AGE", doc.attrib['OID'])
        self.assertEqual("FALSE", doc.attrib['NeedsRetesting'])
        self.assertEqual("FALSE", doc.attrib['BypassDuringMigration'])
        self.assertEqual("TRUE", doc.attrib['AllVariablesInFolders'])
        self.assertEqual("TRUE", doc.attrib['AllVariablesInFields'])
        self.assertEqual("VAROID", doc.attrib['VariableOID'])
        self.assertEqual("FIELDOID", doc.attrib['FieldOID'])
        self.assertEqual("MyForm", doc.attrib['FormOID'])
        self.assertEqual("MyFolder", doc.attrib['FolderOID'])
        self.assertEqual("0", doc.attrib['RecordPosition'])
        self.assertEqual("2", doc.attrib['FormRepeatNumber'])
        self.assertEqual("3", doc.attrib['FolderRepeatNumber'])
        self.assertEqual("MaxBySubject", doc.attrib['LogicalRecordPosition'])

        self.assertEqual("mdsol:DerivationStep", list(doc)[0].tag)

    def test_cannot_accept_any_child(self):
        with self.assertRaises(ValueError):
            MdsolDerivationDef("OID") << object()

    def test_accepts_derivation_step(self):
        tested = MdsolDerivationDef("AGE")
        ds = MdsolDerivationStep(function=StepType.Age)
        tested << ds
        self.assertEqual(ds, tested.derivation_steps[0])

    def test_create_with_valid_lrp(self):
        """We create a function with a valid LRP value"""
        tested = MdsolDerivationDef('TANK',
                                    logical_record_position=LogicalRecordPositionType.Last)
        doc = obj_to_doc(tested)
        self.assertEqual("mdsol:DerivationDef", doc.tag)
        self.assertEqual("Last", doc.attrib['LogicalRecordPosition'])

    def test_create_with_invalid_lrp(self):
        """We create a function with an invalid LRP value"""
        with self.assertRaises(AttributeError) as exc:
            tested = MdsolDerivationDef('TANK',
                                        logical_record_position='Wibble')
        self.assertEqual("Invalid Derivation Def Logical Record Position Wibble",
                         str(exc.exception))


class TestMetaDataVersion(unittest.TestCase):
    """Contains Metadata for study design. Rave only allows one, the spec allows many in an ODM doc"""

    def test_cannot_accept_non_protocol_child(self):
        with self.assertRaises(ValueError):
            MetaDataVersion("OID", "NAME") << object()

    def test_can_accept_protocol_child(self):
        prot = Protocol()
        tested = MetaDataVersion("OID", "NAME")(prot)
        self.assertEqual(prot, tested.protocol)

    def test_can_accept_study_eventdef_child(self):
        sed = StudyEventDef("OID", "Name", False, StudyEventDef.SCHEDULED)
        tested = MetaDataVersion("OID", "NAME")(sed)
        self.assertEqual(sed, tested.study_event_defs[0])

    def test_can_accept_confirmation_message(self):
        cm = MdsolConfirmationMessage("Form saved", lang="eng")
        tested = MetaDataVersion("OID", "NAME")(cm)
        self.assertEqual(cm, tested.confirmation_message)

    def test_can_accept_edit_check_def(self):
        ec = MdsolEditCheckDef("CHECK1")
        tested = MetaDataVersion("OID", "NAME")(ec)
        self.assertEqual(ec, tested.edit_checks[0])

    def test_can_accept_derivation_def(self):
        dd = MdsolDerivationDef("DEV1")
        tested = MetaDataVersion("OID", "NAME")(dd)
        self.assertEqual(dd, tested.derivations[0])

    def test_can_accept_custom_function_def(self):
        cf = MdsolCustomFunctionDef("DEV1", "return true;", language=MdsolCustomFunctionDef.C_SHARP)
        tested = MetaDataVersion("OID", "NAME")(cf)
        self.assertEqual(cf, tested.custom_functions[0])

    def test_builder(self):
        """XML produced"""
        tested = MetaDataVersion("OID", "NAME", description="A description",
                                 primary_formoid="DM",
                                 default_matrix_oid="DEFAULT",
                                 signature_prompt='I confirm.',
                                 delete_existing=True)

        tested << Protocol()
        tested << StudyEventDef("OID", "Name", False, StudyEventDef.SCHEDULED)
        tested << FormDef("FORM_OID", "FORM_Name")
        tested << ItemGroupDef("IG_DEMO", "Demography")
        tested << ItemDef("ID_AGE", "Demography", DataType.Integer, 3)
        tested << CodeList("C_YESNO", "Yes No", DataType.String)
        tested << MdsolLabelDef("LABEL1", "first label")
        tested << MdsolConfirmationMessage("Form has been submitted!")
        tested << MdsolCustomFunctionDef("AlwaysTrue*", "return true;")
        tested << MdsolEditCheckDef("DM01")
        tested << MdsolDerivationDef("AGE")

        doc = obj_to_doc(tested)
        self.assertEqual(doc.tag, "MetaDataVersion")
        self.assertEqual("OID", doc.attrib['OID'])
        self.assertEqual("NAME", doc.attrib['Name'])
        self.assertEqual("A description", doc.attrib['Description'])
        self.assertEqual("DEFAULT", doc.attrib['mdsol:DefaultMatrixOID'])
        self.assertEqual("I confirm.", doc.attrib['mdsol:SignaturePrompt'])
        self.assertEqual("DM", doc.attrib['mdsol:PrimaryFormOID'])
        self.assertEqual("Yes", doc.attrib['mdsol:DeleteExisting'])
        self.assertEqual("Protocol", list(doc)[0].tag)
        self.assertEqual("StudyEventDef", list(doc)[1].tag)
        self.assertEqual("FormDef", list(doc)[2].tag)
        self.assertEqual("ItemGroupDef", list(doc)[3].tag)
        self.assertEqual("ItemDef", list(doc)[4].tag)
        self.assertEqual("CodeList", list(doc)[5].tag)
        self.assertEqual("mdsol:ConfirmationMessage", list(doc)[6].tag)
        self.assertEqual("mdsol:LabelDef", list(doc)[7].tag)
        self.assertEqual("mdsol:EditCheckDef", list(doc)[8].tag)
        self.assertEqual("mdsol:DerivationDef", list(doc)[9].tag)
        self.assertEqual("mdsol:CustomFunctionDef", list(doc)[10].tag)


class TestStudy(unittest.TestCase):
    """Test Study Metadata class"""

    def test_oid(self):
        tested = Study('oid1')
        self.assertEqual('oid1', tested.oid)

    def test_default_project_type(self):
        tested = Study('oid1')
        self.assertEqual('Project', tested.project_type)

    def test_invalid_project_type(self):
        with self.assertRaises(ValueError):
            Study('oid1', project_type='Prijket')

    def test_provided_project_type(self):
        tested = Study('oid1', 'GlobalLibrary Volume')
        self.assertEqual('GlobalLibrary Volume', tested.project_type)

    def test_cannot_accept_itemdata(self):
        tested = Study('oid')
        with self.assertRaises(ValueError):
            tested << ItemData("Field1", "ValueC")

    def test_can_accept_globalvariables(self):
        tested = Study('oid')
        gv = GlobalVariables('MY_TEST_PROJECT')
        tested << gv
        self.assertEqual(gv, tested.global_variables)

    def test_cannot_accept_two_globalvariables(self):
        tested = Study('oid')(GlobalVariables('MY_TEST_PROJECT'))
        with self.assertRaises(ValueError):
            tested << GlobalVariables('Another_one')

    def test_can_accept_basic_definitions(self):
        tested = Study('oid')
        bd = BasicDefinitions()
        tested << bd
        self.assertEqual(bd, tested.basic_definitions)

    def test_cannot_accept_two_basic_definitions(self):
        tested = Study('oid')(BasicDefinitions())
        with self.assertRaises(ValueError):
            tested << BasicDefinitions()

    def test_can_accept_metadata_version(self):
        tested = Study('oid')
        mv = MetaDataVersion('OID', 'Name')
        tested << mv
        self.assertEqual(mv, tested.metadata_version)

    def test_cannot_accept_two_metadata_versions(self):
        tested = Study('oid')(MetaDataVersion('OID1', 'NAME1'))
        with self.assertRaises(ValueError):
            tested << MetaDataVersion('OID2', 'NAME2')

    def test_builder(self):
        """XML produced"""
        tested = Study('oid1', 'GlobalLibrary Volume')(
            GlobalVariables('MY_TEST_PROJECT'),
            BasicDefinitions(),
            MetaDataVersion("OID", "NAME")
        )
        doc = obj_to_doc(tested)
        self.assertEqual(doc.tag, "Study")
        self.assertEqual(doc.attrib['mdsol:ProjectType'], "GlobalLibrary Volume")
        self.assertEqual(list(doc)[0].tag, "GlobalVariables")
        self.assertEqual(list(doc)[1].tag, "BasicDefinitions")
        self.assertEqual(list(doc)[2].tag, "MetaDataVersion")


class TestAlias(unittest.TestCase):

    def test_alias(self):
        """Create an Alias"""
        obj = Alias(context="SDTM", name="DM.BRTHDTC")
        doc = obj_to_doc(obj)
        self.assertEqual('Alias', doc.tag)
        self.assertEqual('SDTM', doc.get('Context'))
        self.assertEqual('DM.BRTHDTC', doc.get('Name'))

    def test_add_alias_to_item(self):
        """Add an Alias to an ItemDef"""
        obj = ItemDef(oid="DM.BRTHDAT", name="Date of Birth", datatype=DataType.Date,
                      date_time_format="dd MMM yyyy")
        obj << Alias(context="SDTM", name="DM.BRTHDTC")
        doc = obj_to_doc(obj)
        self.assertEqual('ItemDef', doc.tag)
        self.assertEqual('Alias', list(doc)[0].tag)

    def test_add_alias_to_protocol(self):
        """Add an Alias to an Protocol"""
        obj = Protocol()
        obj << Alias(context="SDTM", name="DM.BRTHDTC")
        doc = obj_to_doc(obj)
        self.assertEqual('Protocol', doc.tag)
        self.assertEqual('Alias', list(doc)[0].tag)

    def test_add_alias_to_studyevent(self):
        """Add an Alias to an StudyEventDef"""
        obj = StudyEventDef(oid="DM", name="Screening", repeating=False, event_type=StudyEventDef.SCHEDULED)
        obj << Alias(context="MAPPING", name="Screening")
        doc = obj_to_doc(obj)
        self.assertEqual('StudyEventDef', doc.tag)
        self.assertEqual('Alias', list(doc)[0].tag)

    def test_add_alias_to_form(self):
        """Add an Alias to an FormDef"""
        obj = FormDef(oid="DM", name="Demographics", repeating=False)
        obj << Alias(context="MAPPING", name="DEMOG")
        doc = obj_to_doc(obj)
        self.assertEqual('FormDef', doc.tag)
        self.assertEqual('Alias', list(doc)[0].tag)

    def test_add_alias_to_itemgroup(self):
        """Add an Alias to an ItemGroupDef"""
        obj = ItemGroupDef(oid="VS", name="Vitals", repeating=False)
        obj << Alias(context="MAPPING", name="VITALS")
        doc = obj_to_doc(obj)
        self.assertEqual('ItemGroupDef', doc.tag)
        self.assertEqual('Alias', list(doc)[0].tag)

    def test_add_alias_to_codelist(self):
        """Add an Alias to an CodeList"""
        obj = CodeList(oid="VSTEST", name="Vitals Tests", datatype=DataType.String)
        obj << Alias(context="MAPPING", name="VITALS")
        doc = obj_to_doc(obj)
        self.assertEqual('CodeList', doc.tag)
        self.assertEqual('Alias', list(doc)[0].tag)

    def test_add_alias_to_codelistitem(self):
        """Add an Alias to an CodeListItem"""
        obj = CodeListItem(coded_value="SYSBP")
        obj << Alias(context="MAPPING", name="Systolic Blood Pressure")
        doc = obj_to_doc(obj)
        self.assertEqual('CodeListItem', doc.tag)
        self.assertEqual('Alias', list(doc)[0].tag)




# -*- coding: utf-8 -*-

__author__ = 'anewbigging'

import datetime
import hashlib
from lxml import etree
from faker import Factory
from rwslib.extras.rwscmd.odmutils import E_ODM, A_ODM

fake = Factory.create()


def typeof_rave_data(value):
    """Function to duck-type values, not relying on standard Python functions because, for example,
    a string of '1' should be typed as an integer and not as a string or float
    since we're trying to replace like with like when scrambling."""

    # Test if value is a date
    for format in ['%d %b %Y', '%b %Y', '%Y', '%d %m %Y', '%m %Y', '%d/%b/%Y', '%b/%Y', '%d/%m/%Y', '%m/%Y']:
        try:
            datetime.datetime.strptime(value, format)
            if len(value) == 4 and (int(value) < 1900 or int(value) > 2030):
                break
            return ('date', format)
        except ValueError:
            pass
        except TypeError:
            pass

    # Test if value is a time
    for format in ['%H:%M:%S', '%H:%M', '%I:%M:%S', '%I:%M', '%I:%M:%S %p', '%I:%M %p']:
        try:
            datetime.datetime.strptime(value, format)
            return ('time', format)
        except ValueError:
            pass
        except TypeError:
            pass

    # Test if value is a integer
    try:
        if ((isinstance(value, str) and isinstance(int(value), int)) \
                    or isinstance(value, int)):
            return ('int', None)
    except ValueError:
        pass
    except TypeError:
        pass

    # Test if value is a float
    try:
        float(value)
        return ('float', None)
    except ValueError:
        pass
    except TypeError:
        pass

    # If no match on anything else, assume its a string
    return ('string', None)


class Scramble():
    def __init__(self, metadata=None):
        # If initialized with metadata, store relevant formats and lookup information
        if metadata:
            self.metadata = etree.fromstring(metadata)
        else:
            self.metadata = None

    def scramble_int(self, length):
        """Return random integer up to specified number of digits"""
        return str(fake.random_number(length))

    def scramble_float(self, length, sd=0):
        """Return random float in specified format"""
        if sd == 0:
            return str(fake.random_number(length))
        else:
            return str(fake.pyfloat(length - sd, sd, positive=True))

    def scramble_date(self, value, format='%d %b %Y'):
        """Return random date """
        # faker method signature changed
        if value == '':
            # handle the empty string by defaulting to 'now' and 1 year ago
            end_date = 'now'
            start_date = '-1y'
        else:
            # specified end date, and one year prior
            end_date = datetime.datetime.strptime(value, format).date()
            start_date = end_date - datetime.timedelta(days=365)
        fake_date = fake.date_time_between(start_date=start_date,
                                           end_date=end_date).strftime(format).upper()
        return fake_date

    def scramble_time(self, format='%H:%M:%S'):
        """Return random time"""
        return fake.time(pattern=format)

    def scramble_string(self, length):
        """Return random string"""
        return fake.text(length) if length > 5 else ''.join([fake.random_letter() for n in range(0, length)])

    def scramble_value(self, value):
        """Duck-type value and scramble appropriately"""
        try:
            type, format = typeof_rave_data(value)
            if type == 'float':
                i, f = value.split('.')
                return self.scramble_float(len(value) - 1, len(f))
            elif type == 'int':
                return self.scramble_int(len(value))
            elif type == 'date':
                return self.scramble_date(value, format)
            elif type == 'time':
                return self.scramble_time(format)
            elif type == 'string':
                return self.scramble_string(len(value))
            else:
                return value
        except:
            return ""

    def scramble_subjectname(self, value):
        """Scramble subject name with a consistent one-way hash"""
        # md5 will give a consistent obscured value.
        # TODO:  pattern match the subjectname and string replace?
        # TODO:  leave 'New Subject' un-encoded?
        md5 = hashlib.md5(value)
        return md5.hexdigest()

    def scramble_codelist(self, codelist):
        """Return random element from code list"""
        # TODO:  External code lists
        path = ".//{0}[@{1}='{2}']".format(E_ODM.CODELIST.value, A_ODM.OID.value, codelist)
        elem = self.metadata.find(path)
        codes = []
        for c in elem.iter(E_ODM.CODELIST_ITEM.value):
            codes.append(c.get(A_ODM.CODED_VALUE.value))
        for c in elem.iter(E_ODM.ENUMERATED_ITEM.value):
            codes.append(c.get(A_ODM.CODED_VALUE.value))

        return fake.random_element(codes)

    def scramble_itemdata(self, oid, value):
        """If metadata provided, use it to scramble the value based on data type"""
        if self.metadata is not None:
            path = ".//{0}[@{1}='{2}']".format(E_ODM.ITEM_DEF.value, A_ODM.OID.value, oid)
            elem = self.metadata.find(path)
            # for elem in self.metadata.iter(E_ODM.ITEM_DEF.value):
            datatype = elem.get(A_ODM.DATATYPE.value)

            codelist = None
            for el in elem.iter(E_ODM.CODELIST_REF.value):
                codelist = el.get(A_ODM.CODELIST_OID.value)

            length = 1 if not A_ODM.LENGTH in elem else int(elem.get(A_ODM.LENGTH.value))

            if A_ODM.SIGNIFICANT_DIGITS.value in elem.keys():
                sd = elem.get(A_ODM.SIGNIFICANT_DIGITS.value)
            else:
                sd = 0

            if A_ODM.DATETIME_FORMAT.value in elem.keys():
                dt_format = elem.get(A_ODM.DATETIME_FORMAT.value)
                for fmt in [('yyyy', '%Y'), ('MMM', '%b'), ('dd', '%d'), ('HH', '%H'), ('nn', '%M'), ('ss', '%S'),
                            ('-', '')]:
                    dt_format = dt_format.replace(fmt[0], fmt[1])

            if codelist is not None:
                return self.scramble_codelist(codelist)

            elif datatype == 'integer':
                return self.scramble_int(length)

            elif datatype == 'float':
                return self.scramble_float(length, sd)

            elif datatype in ['string', 'text']:
                return self.scramble_string(length)

            elif datatype in ['date', 'datetime']:
                return self.scramble_date(value, dt_format)

            elif datatype in ['time']:
                return self.scramble_time(dt_format)

            else:
                return self.scramble_value(value)

        else:
            return self.scramble_value(value)

    def scramble_query_value(self, value):
        """Return random text for query"""
        return self.scramble_value(value)

    def fill_empty(self, fixed_values, input):
        """Fill in random values for all empty-valued ItemData elements in an ODM document"""
        odm_elements = etree.fromstring(input)

        for v in odm_elements.iter(E_ODM.ITEM_DATA.value):
            if v.get(A_ODM.VALUE.value) == "":
                oid = v.get(A_ODM.ITEM_OID.value)

                if fixed_values is not None and oid in fixed_values:
                    d = fixed_values[oid]
                else:
                    d = self.scramble_itemdata(v.get(A_ODM.ITEM_OID.value), v.get(A_ODM.VALUE.value))

                v.set(A_ODM.VALUE.value, d)
            else:
                # Remove ItemData if it already has a value
                v.getparent().remove(v)

        # Remove empty ItemGroupData elements
        for v in odm_elements.iter(E_ODM.ITEM_GROUP_DATA.value):
            if len(v) == 0:
                v.getparent().remove(v)

        # Remove empty FormData elements
        for v in odm_elements.iter(E_ODM.FORM_DATA.value):
            if len(v) == 0:
                v.getparent().remove(v)

        # Remove empty StudyEventData elements
        for v in odm_elements.iter(E_ODM.STUDY_EVENT_DATA.value):
            if len(v) == 0:
                v.getparent().remove(v)

        return etree.tostring(odm_elements)



# -*- coding: utf-8 -*-
__author__ = 'glow'

from rwslib.builders.clinicaldata import ClinicalData
from rwslib.builders.admindata import AdminData
from rwslib.builders.metadata import Study
from rwslib.builders.common import ODMElement, now_to_iso8601, indent
from rwslib.builders.constants import GranularityType
from xml.etree import cElementTree as ET

import uuid


class ODM(ODMElement):
    """Models the ODM object"""
    FILETYPE_TRANSACTIONAL = 'Transactional'
    FILETYPE_SNAPSHOT = 'Snapshot'

    def __init__(self, originator, description="",
                 creationdatetime=None,
                 fileoid=None,
                 filetype=None,
                 granularity=GranularityType.AllClinicalData,
                 source_system=None,
                 source_system_version=None,
                 respect_unique_file_oid=None):
        """
        :param str originator: The organization that generated the ODM file.
        :param str description: The sender should use the Description attribute to record any information that will 
            help the receiver interpret the document correctly.
        :param str creationdatetime: Time of creation of the file containing the document.
        :param str fileoid: A unique identifier for this file.
        :param str filetype: Snapshot means that the document contains only the current state of the data and metadata 
            it describes, and no transactional history. A Snapshot document may include only one instruction per 
            data point. For clinical data, TransactionType in a Snapshot file must either not be present or be Insert. 
            Transactional means that the document may contain more than one instruction per data point. 
            Use a Transactional document to send both what the current state of the data is, and how it came to be there.
        :param str source_system: The name of the source system that is sending the data.
        :param str source_system_version: The version of the source system that is sending the data.
        :param bool respect_unique_file_oid: Whether to respect the unique file OID constraint.
        """
        self.originator = originator  # Required
        self.description = description
        self.creationdatetime = creationdatetime or now_to_iso8601()
        self.source_system = source_system
        self.source_system_version = source_system_version
        # filetype will always be "Transactional"
        # ODM version will always be 1.3
        # Granularity="SingleSubject"
        # AsOfDateTime always OMITTED (it's optional)
        self.clinical_data = []
        self.study = None
        self.filetype = ODM.FILETYPE_TRANSACTIONAL if filetype is None else ODM.FILETYPE_SNAPSHOT
        self.admindata = None
        # Create unique fileoid if none given
        self.fileoid = str(uuid.uuid4()) if fileoid is None else fileoid
        self._granularity_type = None
        if granularity:
            self.granularity_type = granularity
        self.respect_unique_file_oid = respect_unique_file_oid

    @property
    def granularity_type(self):
        return self._granularity_type

    @granularity_type.setter
    def granularity_type(self, value):
        if not isinstance(value, (GranularityType)):
            raise ValueError("Should be an instance of GranularityType not {}".format(type(value)))
        self._granularity_type = value

    def __lshift__(self, other):
        """Override << operator"""
        if not isinstance(other, (ClinicalData, Study, AdminData)):
            raise ValueError("ODM object can only receive ClinicalData, Study or AdminData object")
        # per the ODM, we can have multiple ClinicalData elements
        self.set_list_attribute(other, ClinicalData, 'clinical_data')
        # per the ODM, we can have multiple Study elements, but owing to the context we restrict it here
        self.set_single_attribute(other, Study, 'study')
        self.set_single_attribute(other, AdminData, 'admindata')

        return other

    def getroot(self):
        """Build XML object, return the root"""
        builder = ET.TreeBuilder()
        self.build(builder)
        return builder.close()

    def build(self, builder):
        """Build XML object, return the root, this is a copy for consistency and testing"""
        params = dict(ODMVersion="1.3",
                      FileType=self.filetype,
                      CreationDateTime=self.creationdatetime,
                      Originator=self.originator,
                      FileOID=self.fileoid,
                      xmlns="http://www.cdisc.org/ns/odm/v1.3",
                      )
        if self.granularity_type:
            params['Granularity'] = self.granularity_type.value
        if self.source_system:
            params['SourceSystem'] = self.source_system

        if self.source_system_version:
            params['SourceSystemVersion'] = self.source_system_version
        params['xmlns:mdsol'] = "http://www.mdsol.com/ns/odm/metadata"

        if self.description:
            params['Description'] = self.description

        if self.respect_unique_file_oid is not None:
            params['RespectUniqueFileOID'] = "Yes" if self.respect_unique_file_oid else "No"

        builder.start("ODM", params)

        # Ask the children
        if self.study is not None:
            self.study.build(builder)

        if self.clinical_data:
            for clinical_data in self.clinical_data:
                clinical_data.build(builder)

        if self.admindata is not None:
            self.admindata.build(builder)

        builder.end("ODM")
        return builder.close()

    def __str__(self):
        doc = self.getroot()
        indent(doc)
        header = '<?xml version="1.0" encoding="utf-8" ?>\n'
        return header + ET.tostring(doc, encoding='utf-8').decode('utf-8')



# -*- coding: utf-8 -*-
__author__ = 'isparks'

import enum


class DataType(enum.Enum):
    """ODM Data Types"""
    Text = 'text'
    Integer = 'integer'
    Float = 'float'
    Date = 'date'
    DateTime = 'datetime'
    Time = 'time'
    String = 'string'  # Used only by codelists


class QueryStatusType(enum.Enum):
    """
    MdsolQuery action type
    Applies to a :class:`MdsolQuery`
    """
    Open = "Open"
    Cancelled = "Cancelled"
    Answered = "Answered"
    Forwarded = "Forwarded"
    Closed = "Closed"


class StepType(enum.Enum):
    """
    Edit/Derivation step types
    Applies to a :class:`MdsolCheckStep`, :class:`MdsolDerivationStep`
    """
    CustomFunction = "CustomFunction"
    IsEmpty = "IsEmpty"
    IsNotEmpty = "IsNotEmpty"
    Contains = "Contains"
    StartsWith = "StartsWith"
    IsLessThan = "IsLessThan"
    IsLessThanOrEqualTo = "IsLessThanOrEqualTo"
    IsGreaterThan = "IsGreaterThan"
    IsGreaterThanOrEqualTo = "IsGreaterThanOrEqualTo"
    IsEqualTo = "IsEqualTo"
    IsNonConformant = "IsNonConformant"
    IsNotEqualTo = "IsNotEqualTo"
    InLocalLabRange = "InLocalLabRange"
    LengthIsLessThan = "LengthIsLessThan"
    LengthIsLessThanOrEqualTo = "LengthIsLessThanOrEqualTo"
    LengthIsGreaterThan = "LengthIsGreaterThan"
    LengthIsGreaterThanOrEqualTo = "LengthIsGreaterThanOrEqualTo"
    LengthIsEqualTo = "LengthIsEqualTo"
    Or = "Or"
    And = "And"
    Not = "Not"
    Now = "Now"
    IsPresent = "IsPresent"
    IsActive = "IsActive"
    Add = "Add"
    Subtract = "Subtract"
    Multiply = "Multiply"
    Divide = "Divide"
    AddDay = "AddDay"
    AddMonth = "AddMonth"
    AddYear = "AddYear"
    AddSec = "AddSec"
    AddMin = "AddMin"
    AddHour = "AddHour"
    DaySpan = "DaySpan"
    TimeSpan = "TimeSpan"
    Age = "Age"
    StringAdd = "StringAdd"
    Space = "Space"


ALL_STEPS = [StepType.CustomFunction,
             StepType.IsEmpty,
             StepType.IsNotEmpty,
             StepType.Contains,
             StepType.StartsWith,
             StepType.IsLessThan,
             StepType.IsLessThanOrEqualTo,
             StepType.IsGreaterThan,
             StepType.IsGreaterThanOrEqualTo,
             StepType.IsEqualTo,
             StepType.IsNonConformant,
             StepType.IsNotEqualTo,
             StepType.InLocalLabRange,
             StepType.LengthIsLessThan,
             StepType.LengthIsLessThanOrEqualTo,
             StepType.LengthIsGreaterThan,
             StepType.LengthIsGreaterThanOrEqualTo,
             StepType.LengthIsEqualTo,
             StepType.Or,
             StepType.And,
             StepType.Not,
             StepType.Now,
             StepType.IsPresent,
             StepType.IsActive,
             StepType.Add,
             StepType.Subtract,
             StepType.Multiply,
             StepType.Divide,
             StepType.AddDay,
             StepType.AddMonth,
             StepType.AddYear,
             StepType.AddSec,
             StepType.AddMin,
             StepType.AddHour,
             StepType.DaySpan,
             StepType.TimeSpan,
             StepType.Age,
             StepType.StringAdd]

# Note: Missing 2015 additions to edit check step functions.


VALID_DERIVATION_STEPS = [
    StepType.Age,
    StepType.Subtract,
    StepType.Multiply,
    StepType.Divide,
    StepType.AddDay,
    StepType.AddMonth,
    StepType.AddYear,
    StepType.AddSec,
    StepType.AddMin,
    StepType.AddHour,
    StepType.DaySpan,
    StepType.TimeSpan,
    StepType.Now,
    StepType.StringAdd,
    StepType.CustomFunction,
    StepType.Space,
    StepType.Add
]


class ActionType(enum.Enum):
    """
    CheckAction types
    Applies to a :class:`CheckAction`
    """
    OpenQuery = "OpenQuery"
    RequireReview = "RequireReview"
    RequireVerification = "RequireVerification"
    AddComment = "AddComment"
    AddDeviation = "AddDeviation"
    CustomFunction = "CustomFunction"
    PlaceSticky = "PlaceSticky"
    AddForm = "AddForm"
    AddMatrix = "AddMatrix"
    MrgMatrix = "MrgMatrix"
    OldMrgMatrix = "OldMrgMatrix"
    SetNonconformant = "SetNonconformant"
    SendMessage = "SendMessage"
    SetDataPoint = "SetDataPoint"
    SetTimeZero = "SetTimeZero"
    SetTimeForward = "SetTimeForward"
    SetSubjectStatus = "SetSubjectStatus"
    SetSubjectName = "SetSubjectName"
    UpdateFormName = "UpdateFormName"
    UpdateFolderName = "UpdateFolderName"
    SetRecordDate = "SetRecordDate"
    SetDataPageDate = "SetDataPageDate"
    SetInstanceDate = "SetInstanceDate"
    SetSubjectDate = "SetSubjectDate"
    SetDataPointVisible = "SetDataPointVisible"
    SetSecondarySubjectName = "SetSecondarySubjectName"
    SetFormRequiresSignature = "SetFormRequiresSignature"
    SetFolderRequiresSignature = "SetFolderRequiresSignature"
    SetSubjectRequiresSignature = "SetSubjectRequiresSignature"
    SetDynamicSearchList = "SetDynamicSearchList"


ALL_ACTIONS = [
    ActionType.OpenQuery,
    ActionType.RequireReview,
    ActionType.RequireVerification,
    ActionType.AddComment,
    ActionType.AddDeviation,
    ActionType.CustomFunction,
    ActionType.PlaceSticky,
    ActionType.AddForm,
    ActionType.AddMatrix,
    ActionType.MrgMatrix,
    ActionType.OldMrgMatrix,
    ActionType.SetNonconformant,
    ActionType.SendMessage,
    ActionType.SetDataPoint,
    ActionType.SetTimeZero,
    ActionType.SetTimeForward,
    ActionType.SetSubjectStatus,
    ActionType.SetSubjectName,
    ActionType.UpdateFormName,
    ActionType.UpdateFolderName,
    ActionType.SetRecordDate,
    ActionType.SetDataPageDate,
    ActionType.SetInstanceDate,
    ActionType.SetSubjectDate,
    ActionType.SetDataPointVisible,
    ActionType.SetSecondarySubjectName,
    ActionType.SetFormRequiresSignature,
    ActionType.SetFolderRequiresSignature,
    ActionType.SetSubjectRequiresSignature,
    ActionType.SetDynamicSearchList
]


class RangeCheckComparatorType(enum.Enum):
    """
    Range Check Enumeration
    Applies to a :class:`RangeCheck`
    """
    LessThanEqualTo = 'LE'
    GreaterThanEqualTo = 'GE'


class RangeCheckType(enum.Enum):
    """
    Range Check Type Enumeration
    Applies to a :class:`RangeCheck`
    """
    Soft = 'Soft'
    Hard = 'Hard'


class ControlType(enum.Enum):
    """
    CRF Control Type Enumeration
    Applies to a :class:`ItemDef`
    """
    CheckBox = 'CheckBox'
    Text = 'Text'
    DateTime = 'DateTime'
    DropDownList = 'DropDownList'
    SearchList = 'SearchList'
    RadioButton = 'RadioButton'
    RadioButtonVertical = 'RadioButton (Vertical)'
    FileUpload = 'File Upload'
    LongText = 'LongText'
    SignaturePage = 'Signature page'
    SignatureFolder = 'Signature folder'
    SignatureSubject = 'Signature subject'


class LogicalRecordPositionType(enum.Enum):
    """
    Logical Record Position (LRP) Type Enumeration
    Applies to a :class:`MdsolDerivationDef`, :class:`MdsolDerivationStep`, :class:`MdsolCheckStep`
    """
    MaxBySubject = 'MaxBySubject'
    MaxByInstance = 'MaxByInstance'
    MaxByDataPage = 'MaxByDataPage'
    Last = 'Last'
    Next = 'Next'
    Previous = 'Previous'
    First = 'First'
    MinByDataPage = 'MinByDataPage'
    MinByInstance = "MinByInstance"
    MinBySubject = 'MinBySubject'


class ProtocolDeviationStatus(enum.Enum):
    """
    Protocol Deviation Status Enumeration
    Applies to a :class:`MdsolProtocolDeviation`
    """
    Open = "Open"
    Removed = "Removed"


LOGICAL_RECORD_POSITIONS = [
    LogicalRecordPositionType.MaxBySubject,
    LogicalRecordPositionType.MaxBySubject,
    LogicalRecordPositionType.MaxByInstance,
    LogicalRecordPositionType.MaxByDataPage,
    LogicalRecordPositionType.Last,
    LogicalRecordPositionType.Next,
    LogicalRecordPositionType.Previous,
    LogicalRecordPositionType.First,
    LogicalRecordPositionType.MinByDataPage,
    LogicalRecordPositionType.MinByInstance,
    LogicalRecordPositionType.MinBySubject
]


class LocationType(enum.Enum):
    """
    Location Type Enumeration
    Applies to a :class:`Location`
    """
    Sponsor = 'Sponsor'
    Site = 'Site'
    CRO = 'CRO'
    Lab = 'Lab'
    Other = 'Other'


class LabType(enum.Enum):
    """
    Type of LaboratoryType
    Applies to a :class:`FormData`
    """
    Local = 'Local'
    Central = 'Central'


class UserType(enum.Enum):
    """
    User Type Enumeration
    Applies to a :class:`User`
    """
    Sponsor = 'Sponsor'
    Investigator = 'Investigator'
    Lab = 'Lab'
    Other = 'Other'


class GranularityType(enum.Enum):
    """
    ODM Granularity Type Enumeration
    Applies to a :class:`ODM`
    """
    All = 'All'
    Metadata = 'Metadata'
    AdminData = 'AdminData'
    ReferenceData = 'ReferenceData'
    AllClinicalData = 'AllClinicalData'
    SingleSite = 'SingleSite'
    SingleSubject = 'SingleSubject'


class LocationOIDType(enum.Enum):
    """
    Location OID Type Enumeration
    Applies to a :class:`SiteRef`
    """
    SiteUUID = 'SiteUUID'
    SiteNumber = 'SiteNumber'
    StudyEnvSiteNumber = 'StudyEnvSiteNumber'




__author__ = 'isparks'
import unittest

from rwslib.extras.local_cv import SQLLiteDBAdapter, LocalCVBuilder

metadata = """projectname,viewname,ordinal,varname,vartype,varlength,varformat,varlabel
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","1","userid","num","8","10.","Internal id for the user"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","2","projectid","num","8","10.","projectid"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","3","project","char","255","$255.","project"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","4","studyid","num","8","10.","Internal id for the study"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","5","environmentName","char","20","$20.","Environment"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","6","subjectId","num","8","10.","Internal id for the subject"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","7","StudySiteId","num","8","10.","Internal id for study site"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","8","Subject","char","50","$50.","Subject name or identifier"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","9","siteid","num","8","10.","Internal id for the site"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","10","Site","char","255","$255.","Site name"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","11","SiteNumber","char","50","$50.","SiteNumber"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","12","SiteGroup","char","40","$40.","SiteGroup"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","13","instanceId","num","8","10.","Internal id for the instance"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","14","InstanceName","char","255","$255.","Folder instance name"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","15","InstanceRepeatNumber","num","8","10.","InstanceRepeatNumber"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","16","folderid","num","8","10.","Internal id for the folder"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","17","Folder","char","50","$50.","Folder OID"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","18","FolderName","char","255","$255.","Folder name"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","19","FolderSeq","num","8","12.1","Folder sequence number"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","20","TargetDays","num","8","10.","Target days from study start"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","21","DataPageId","num","8","10.","Internal id for data page"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","22","DataPageName","char","255","$255.","eCRF page name"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","23","PageRepeatNumber","num","8","10.","Sequence number of eCRF page in folder"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","24","RecordDate","num","8","datetime22.3","Clinical date of record (ex: visit date)"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","25","RecordId","num","8","10.","Internal id for the record"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","26","recordposition","num","8","10.","Record number"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","27","RecordActive","num","8","1.","Is record active"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","28","SaveTs","num","8","datetime22.3","Timestamp of last save in clinical views"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","29","MinCreated","num","8","datetime22.3","Earliest data creation time"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","30","MaxUpdated","num","8","datetime22.3","Latest data update time"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","31","SUBID","char","10","$10.","SUBID"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","32","BIRTHDT","num","8","datetime22.3","BIRTHDT"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","33","BIRTHDT_RAW","char","11","$11.","BIRTHDT(Character)"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","34","BIRTHDT_INT","num","8","datetime22.3","BIRTHDTInterpolated"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","35","BIRTHDT_YYYY","num","8","4.","BIRTHDTYear"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","36","BIRTHDT_MM","num","8","2.","BIRTHDTMonth"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL","37","BIRTHDT_DD","num","8","2.","BIRTHDTDay"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","1","userid","num","8","10.","Internal id for the user"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","2","projectid","num","8","10.","projectid"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","3","project","char","255","$255.","project"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","4","studyid","num","8","10.","Internal id for the study"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","5","environmentName","char","20","$20.","Environment"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","6","subjectId","num","8","10.","Internal id for the subject"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","7","StudySiteId","num","8","10.","Internal id for study site"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","8","Subject","char","50","$50.","Subject name or identifier"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","9","siteid","num","8","10.","Internal id for the site"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","10","Site","char","255","$255.","Site name"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","11","SiteNumber","char","50","$50.","SiteNumber"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","12","SiteGroup","char","40","$40.","SiteGroup"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","13","instanceId","num","8","10.","Internal id for the instance"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","14","InstanceName","char","255","$255.","Folder instance name"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","15","InstanceRepeatNumber","num","8","10.","InstanceRepeatNumber"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","16","folderid","num","8","10.","Internal id for the folder"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","17","Folder","char","50","$50.","Folder OID"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","18","FolderName","char","255","$255.","Folder name"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","19","FolderSeq","num","8","12.1","Folder sequence number"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","20","TargetDays","num","8","10.","Target days from study start"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","21","DataPageId","num","8","10.","Internal id for data page"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","22","DataPageName","char","255","$255.","eCRF page name"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","23","PageRepeatNumber","num","8","10.","Sequence number of eCRF page in folder"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","24","RecordDate","num","8","datetime22.3","Clinical date of record (ex: visit date)"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","25","RecordId","num","8","10.","Internal id for the record"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","26","recordposition","num","8","10.","Record number"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","27","RecordActive","num","8","1.","Is record active"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","28","SaveTs","num","8","datetime22.3","Timestamp of last save in clinical views"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","29","MinCreated","num","8","datetime22.3","Earliest data creation time"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","30","MaxUpdated","num","8","datetime22.3","Latest data update time"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","31","SUBID","char","10","$10.","SUBID"
"SIMPLESTUDY","V_SIMPLESTUDY_ENROL_RAW","32","BIRTHDT","char","11","$11.","BIRTHDT"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","1","userid","num","8","10.","Internal id for the user"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","2","projectid","num","8","10.","projectid"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","3","project","char","255","$255.","project"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","4","studyid","num","8","10.","Internal id for the study"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","5","environmentName","char","20","$20.","Environment"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","6","subjectId","num","8","10.","Internal id for the subject"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","7","StudySiteId","num","8","10.","Internal id for study site"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","8","Subject","char","50","$50.","Subject name or identifier"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","9","siteid","num","8","10.","Internal id for the site"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","10","Site","char","255","$255.","Site name"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","11","SiteNumber","char","50","$50.","SiteNumber"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","12","SiteGroup","char","40","$40.","SiteGroup"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","13","instanceId","num","8","10.","Internal id for the instance"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","14","InstanceName","char","255","$255.","Folder instance name"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","15","InstanceRepeatNumber","num","8","10.","InstanceRepeatNumber"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","16","folderid","num","8","10.","Internal id for the folder"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","17","Folder","char","50","$50.","Folder OID"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","18","FolderName","char","255","$255.","Folder name"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","19","FolderSeq","num","8","12.1","Folder sequence number"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","20","TargetDays","num","8","10.","Target days from study start"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","21","DataPageId","num","8","10.","Internal id for data page"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","22","DataPageName","char","255","$255.","eCRF page name"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","23","PageRepeatNumber","num","8","10.","Sequence number of eCRF page in folder"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","24","RecordDate","num","8","datetime22.3","Clinical date of record (ex: visit date)"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","25","recordposition","num","8","10.","Record number"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","26","DatapointActive","num","8","1.","DatapointActive"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","27","SaveTs","num","8","datetime22.3","Timestamp of last save in clinical views"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","28","DataPointId","num","8","10.","DataPointId"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","29","RecordId","num","8","10.","Internal id for the record"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","30","FormId","num","8","10.","FormId"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","31","FieldId","num","8","10.","FieldId"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","32","AnalyteId","num","8","10.","AnalyteId"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","33","LabId","num","8","10.","LabId"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","34","AnalyteRangeId","num","8","10.","AnalyteRangeId"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","35","DataDictionaryId","num","8","10.","DataDictionaryId"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","36","LabUnitId","num","8","10.","LabUnitId"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","37","LabStandardGroupId","num","8","10.","LabStandardGroupId"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","38","LabUnitConversionId","num","8","10.","LabUnitConversionId"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","39","StandardLabUnitId","num","8","10.","StandardLabUnitId"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","40","ReferenceLabId","num","8","10.","ReferenceLabId"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","41","ReferenceRangeId","num","8","10.","ReferenceRangeId"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","42","ReferenceDataDictionaryId","num","8","10.","ReferenceDataDictionaryId"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","43","AlertLabId","num","8","10.","AlertLabId"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","44","AlertRangeId","num","8","10.","AlertRangeId"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","45","AlertDataDictionaryId","num","8","10.","AlertDataDictionaryId"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","46","Created","num","8","datetime22.3","Created"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","47","Updated","num","8","datetime22.3","Updated"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","48","Form","char","50","$50.","Form"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","49","FormName","char","255","$255.","FormName"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","50","fieldOrdinal","num","8","10.","fieldOrdinal"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","51","AnalyteName","char","255","$255.","AnalyteName"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","52","AnalyteValue","char","2000","$2000.","AnalyteValue"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","53","NumericValue","num","8","16.7","NumericValue"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","54","CodedValue","char","255","$255.","CodedValue"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","55","LabName","char","255","$255.","LabName"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","56","RangesApproved","num","8","1.","RangesApproved"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","57","LabDictionary","char","255","$255.","LabDictionary"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","58","LabLow","num","8","16.7","LabLow"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","59","LabHigh","num","8","16.7","LabHigh"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","60","LabUnits","char","255","$255.","LabUnits"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","61","LabFlag","char","1","$1.","LabFlag"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","62","LabComments","char","255","$255.","LabComments"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","63","StdGroup","char","255","$255.","StdGroup"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","64","StdValue","num","8","16.7","StdValue"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","65","StdLow","num","8","16.7","StdLow"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","66","StdHigh","num","8","16.7","StdHigh"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","67","StdUnits","char","255","$255.","StdUnits"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","68","RefLab","char","255","$255.","RefLab"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","69","RefDictionary","char","255","$255.","RefDictionary"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","70","RefLow","num","8","16.7","RefLow"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","71","RefHigh","num","8","16.7","RefHigh"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","72","RefFlag","char","1","$1.","RefFlag"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","73","RefComments","char","255","$255.","RefComments"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","74","AlertLab","char","255","$255.","AlertLab"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","75","AlertDictionary","char","255","$255.","AlertDictionary"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","76","AlertLow","num","8","16.7","AlertLow"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","77","AlertHigh","num","8","16.7","AlertHigh"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","78","AlertFlag","char","1","$1.","AlertFlag"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","79","AlertComments","char","255","$255.","AlertComments"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","80","ClinSigStatus","num","8","10.","ClinSigStatus"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","81","ClinSigValue","char","50","$50.","ClinSigValue"
"SIMPLESTUDY","V_SIMPLESTUDY_Lab","82","ClinSigComment","char","255","$255.","ClinSigComment"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","1","userid","num","8","10.","Internal id for the user"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","2","projectid","num","8","10.","projectid"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","3","project","char","255","$255.","project"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","4","studyid","num","8","10.","Internal id for the study"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","5","environmentName","char","20","$20.","Environment"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","6","subjectId","num","8","10.","Internal id for the subject"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","7","StudySiteId","num","8","10.","Internal id for study site"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","8","Subject","char","50","$50.","Subject name or identifier"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","9","siteid","num","8","10.","Internal id for the site"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","10","Site","char","255","$255.","Site name"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","11","SiteNumber","char","50","$50.","SiteNumber"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","12","SiteGroup","char","40","$40.","SiteGroup"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","13","instanceId","num","8","10.","Internal id for the instance"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","14","InstanceName","char","255","$255.","Folder instance name"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","15","InstanceRepeatNumber","num","8","10.","InstanceRepeatNumber"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","16","folderid","num","8","10.","Internal id for the folder"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","17","Folder","char","50","$50.","Folder OID"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","18","FolderName","char","255","$255.","Folder name"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","19","FolderSeq","num","8","12.1","Folder sequence number"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","20","TargetDays","num","8","10.","Target days from study start"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","21","DataPageId","num","8","10.","Internal id for data page"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","22","DataPageName","char","255","$255.","eCRF page name"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","23","PageRepeatNumber","num","8","10.","Sequence number of eCRF page in folder"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","24","RecordDate","num","8","datetime22.3","Clinical date of record (ex: visit date)"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","25","RecordId","num","8","10.","Internal id for the record"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","26","recordposition","num","8","10.","Record number"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","27","RecordActive","num","8","1.","Is record active"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","28","SaveTs","num","8","datetime22.3","Timestamp of last save in clinical views"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","29","MinCreated","num","8","datetime22.3","Earliest data creation time"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","30","MaxUpdated","num","8","datetime22.3","Latest data update time"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","31","VDAT","num","8","datetime22.3","VDAT"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","32","VDAT_RAW","char","11","$11.","VDAT(Character)"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","33","VDAT_INT","num","8","datetime22.3","VDATInterpolated"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","34","VDAT_YYYY","num","8","4.","VDATYear"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","35","VDAT_MM","num","8","2.","VDATMonth"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","36","VDAT_DD","num","8","2.","VDATDay"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","37","WEIGHT_KG","num","8","5.1","WEIGHT_KG"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","38","WEIGHT_KG_RAW","char","6","$6.","WEIGHT_KG(Character)"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","39","HEIGHT_CM","num","8","5.1","HEIGHT_CM"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL","40","HEIGHT_CM_RAW","char","6","$6.","HEIGHT_CM(Character)"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","1","userid","num","8","10.","Internal id for the user"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","2","projectid","num","8","10.","projectid"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","3","project","char","255","$255.","project"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","4","studyid","num","8","10.","Internal id for the study"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","5","environmentName","char","20","$20.","Environment"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","6","subjectId","num","8","10.","Internal id for the subject"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","7","StudySiteId","num","8","10.","Internal id for study site"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","8","Subject","char","50","$50.","Subject name or identifier"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","9","siteid","num","8","10.","Internal id for the site"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","10","Site","char","255","$255.","Site name"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","11","SiteNumber","char","50","$50.","SiteNumber"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","12","SiteGroup","char","40","$40.","SiteGroup"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","13","instanceId","num","8","10.","Internal id for the instance"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","14","InstanceName","char","255","$255.","Folder instance name"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","15","InstanceRepeatNumber","num","8","10.","InstanceRepeatNumber"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","16","folderid","num","8","10.","Internal id for the folder"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","17","Folder","char","50","$50.","Folder OID"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","18","FolderName","char","255","$255.","Folder name"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","19","FolderSeq","num","8","12.1","Folder sequence number"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","20","TargetDays","num","8","10.","Target days from study start"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","21","DataPageId","num","8","10.","Internal id for data page"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","22","DataPageName","char","255","$255.","eCRF page name"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","23","PageRepeatNumber","num","8","10.","Sequence number of eCRF page in folder"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","24","RecordDate","num","8","datetime22.3","Clinical date of record (ex: visit date)"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","25","RecordId","num","8","10.","Internal id for the record"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","26","recordposition","num","8","10.","Record number"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","27","RecordActive","num","8","1.","Is record active"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","28","SaveTs","num","8","datetime22.3","Timestamp of last save in clinical views"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","29","MinCreated","num","8","datetime22.3","Earliest data creation time"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","30","MaxUpdated","num","8","datetime22.3","Latest data update time"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","31","VDAT","char","11","$11.","VDAT"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","32","WEIGHT_KG","char","6","$6.","WEIGHT_KG"
"SIMPLESTUDY","V_SIMPLESTUDY_VITAL_RAW","33","HEIGHT_CM","char","6","$6.","HEIGHT_CM"
EOF
"""

enrol_data = """userid,projectid,project,studyid,environmentName,subjectId,StudySiteId,Subject,siteid,Site,SiteNumber,SiteGroup,instanceId,InstanceName,InstanceRepeatNumber,folderid,Folder,FolderName,FolderSeq,TargetDays,DataPageId,DataPageName,PageRepeatNumber,RecordDate,RecordId,recordposition,RecordActive,SaveTs,MinCreated,MaxUpdated,SUBID,BIRTHDT,BIRTHDT_RAW,BIRTHDT_INT,BIRTHDT_YYYY,BIRTHDT_MM,BIRTHDT_DD
"457","85","SIMPLESTUDY","95","TEST","32112","143","1","120","TESTSITE","TESTSITE","World","","","","","","","0.0","","662503","Enrol","0","","1346660","0","1","2013-10-07T15:50:33","2013-06-24T09:52:09","2013-06-24T09:52:10","1","1973-06-26T00:00:00","1973 Jun 26","1973-06-26T00:00:00","1973","6","26"
"457","85","SIMPLESTUDY","95","TEST","32113","143","2","120","TESTSITE","TESTSITE","World","","","","","","","0.0","","662505","Enrol","0","","1346662","0","1","2013-10-07T15:50:33","2013-06-24T09:52:11","2013-06-24T09:52:11","2","1971-10-13T00:00:00","1971 Oct 13","1971-10-13T00:00:00","1971","10","13"
"457","85","SIMPLESTUDY","95","TEST","32114","143","3","120","TESTSITE","TESTSITE","World","","","","","","","0.0","","662507","Enrol","0","","1346664","0","1","2013-10-07T15:50:33","2013-06-24T09:52:12","2013-10-07T15:50:18","3","1953-08-10T00:00:00","1953 Aug 10","1953-08-10T00:00:00","1953","8","10"
"457","85","SIMPLESTUDY","95","TEST","32115","143","4","120","TESTSITE","TESTSITE","World","","","","","","","0.0","","662509","Enrol","0","","1346666","0","1","2013-10-07T15:50:33","2013-06-24T09:52:13","2013-06-24T09:52:13","4","1975-02-08T00:00:00","1975 Feb 08","1975-02-08T00:00:00","1975","2","8"
"457","85","SIMPLESTUDY","95","TEST","32116","143","5","120","TESTSITE","TESTSITE","World","","","","","","","0.0","","662511","Enrol","0","","1346668","0","1","2013-10-07T15:50:33","2013-06-24T09:52:15","2013-06-24T09:52:15","5","1977-10-23T00:00:00","1977 Oct 23","1977-10-23T00:00:00","1977","10","23"
"457","85","SIMPLESTUDY","95","TEST","32117","143","6","120","TESTSITE","TESTSITE","World","","","","","","","0.0","","662513","Enrol","0","","1346670","0","1","2013-10-07T15:50:33","2013-06-24T09:52:16","2013-06-24T09:52:16","6","1970-04-18T00:00:00","1970 Apr 18","1970-04-18T00:00:00","1970","4","18"
"457","85","SIMPLESTUDY","95","TEST","32118","143","7","120","TESTSITE","TESTSITE","World","","","","","","","0.0","","662515","Enrol","0","","1346672","0","1","2013-10-07T15:50:33","2013-06-24T09:52:17","2013-06-24T09:52:17","7","1957-10-14T00:00:00","1957 Oct 14","1957-10-14T00:00:00","1957","10","14"
"457","85","SIMPLESTUDY","95","TEST","32119","143","8","120","TESTSITE","TESTSITE","World","","","","","","","0.0","","662517","Enrol","0","","1346674","0","1","2013-10-07T15:50:33","2013-06-24T09:52:18","2013-10-07T15:16:29","8","1973-01-18T00:00:00","1973 Jan 18","1973-01-18T00:00:00","1973","1","18"
"457","85","SIMPLESTUDY","95","TEST","32120","143","9","120","TESTSITE","TESTSITE","World","","","","","","","0.0","","662519","Enrol","0","","1346676","0","1","2013-10-07T15:50:33","2013-06-24T09:52:19","2013-06-24T09:52:19","9","1974-03-04T00:00:00","1974 Mar 04","1974-03-04T00:00:00","1974","3","4"
"457","85","SIMPLESTUDY","95","TEST","32121","143","10","120","TESTSITE","TESTSITE","World","","","","","","","0.0","","662521","Enrol","0","","1346678","0","1","2013-10-07T15:50:33","2013-06-24T09:52:20","2013-06-24T09:52:20","10","1964-03-03T00:00:00","1964 Mar 03","1964-03-03T00:00:00","1964","3","3"
"457","85","SIMPLESTUDY","95","TEST","34128","147","012","123","TESTSITE2","TESTSITE2","World","","","","","","","0.0","","705383","Enrol","0","","1432815","0","1","2013-10-07T15:50:33","2013-10-03T13:18:48","2013-10-07T15:15:45","012","1973-06-26T00:00:00","1973 Jun 26","1973-06-26T00:00:00","1973","6","26"
EOF
"""

class FakeCursor(object):
    def __init__(self):
        self.sql = []

    def __call__(self):
        return self

    def execute(self, *args, **kwargs):
        self.sql.append((args, kwargs,))

    def executemany(self, stmt, rows):
        """Simple fake for unit tests"""
        for row in rows:
            self.execute(stmt, row)

class FakeConn(object):
    """Fake daabase connection."""
    def __init__(self):
        self.cursor = FakeCursor()

    def commit(self):
        pass

class TestRWSCSVReader(unittest.TestCase):
    def setUp(self):
        fake_db = FakeConn()
        self.tested = SQLLiteDBAdapter(fake_db)
        self.tested._setDatasets(metadata)

    def test_dataset(self):
        self.assertEqual("V_SIMPLESTUDY_ENROL" in self.tested.datasets, True)

    def test_cols(self):
        self.assertEqual(self.tested.datasets["V_SIMPLESTUDY_ENROL"][0]["ordinal"], "1")
        self.assertEqual(self.tested.datasets["V_SIMPLESTUDY_ENROL"][1]["varname"], "projectid")
        self.assertEqual(self.tested.datasets["V_SIMPLESTUDY_ENROL"][2]["vartype"], "char")
        self.assertEqual(self.tested.datasets["V_SIMPLESTUDY_ENROL"][3]["varformat"], "10.")
        self.assertEqual(self.tested.datasets["V_SIMPLESTUDY_ENROL"][4]["varlabel"], "Environment")

    def test_table_gen(self):
        "test generating tables"
        sql = self.tested._generateDDL()
        self.assertEqual(sql[0],'drop table if exists V_SIMPLESTUDY_ENROL;')
        self.assertEqual(sql[1][:34],u'CREATE TABLE V_SIMPLESTUDY_ENROL (')
        self.assertEqual(True, 'userid NUMERIC' in sql[1])
        self.assertEqual(True, 'projectid NUMERIC' in sql[1])
        self.assertEqual(True, 'project TEXT' in sql[1])


    def test_insert_gen(self):
        """Test generation of insert statements"""
        self.tested.processFormData(enrol_data, 'V_SIMPLESTUDY_ENROL')

        first_line = self.tested.conn.cursor.sql[0][0]
        first_statement = first_line[0]
        first_values = first_line[1]

        #Check sql construction
        self.assertEqual(first_statement.startswith('INSERT INTO V_SIMPLESTUDY_ENROL (userid,projectid,project'), True)
        self.assertEqual('SUBID,BIRTHDT,BIRTHDT_RAW,BIRTHDT_INT,BIRTHDT_YYYY,BIRTHDT_MM,BIRTHDT_DD)' in first_statement, True)

        #Check values
        self.assertEqual('457',first_values[0]) #First value
        self.assertEqual('26',first_values[-1]) #Last value


class TestSQLDataType(unittest.TestCase):
    def test_sql_typefor(self):

        sqlite_db = None
        tested = SQLLiteDBAdapter(sqlite_db)
        self.assertEqual(tested.getSQLDataType("num"), "NUMERIC")
        self.assertEqual(tested.getSQLDataType("char"), "TEXT")

class TestNameTypeFromViewname(unittest.TestCase):
    def test_name_type_from_viewname(self):
        self.assertEqual(LocalCVBuilder.name_type_from_viewname("V_SIMPLESTUDY_ENROL")[0],"ENROL")
        self.assertEqual(LocalCVBuilder.name_type_from_viewname("V_SIMPLESTUDY_ENROL")[1],"REGULAR")
        self.assertEqual(LocalCVBuilder.name_type_from_viewname("V_SIMPLESTUDY_ENROL_RAW")[0],"ENROL")
        self.assertEqual(LocalCVBuilder.name_type_from_viewname("V_SIMPLESTUDY_ENROL_RAW")[1],"RAW")
        self.assertEqual(LocalCVBuilder.name_type_from_viewname("V_SIMPLESTUDY_Lab")[0],"Lab")
        self.assertEqual(LocalCVBuilder.name_type_from_viewname("V_SIMPLESTUDY_Lab")[1],"REGULAR")
        self.assertEqual(LocalCVBuilder.name_type_from_viewname("V_SIMPLESTUDY_Lab_")[1],"REGULAR")

if __name__ == '__main__':
    unittest.main()
